#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017-2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Tests for the CF writer."""

import datetime as dt
import os
import tempfile
import warnings
from contextlib import nullcontext

import dask.array as da
import numpy as np
import pytest
import xarray as xr
from packaging.version import Version

from satpy import Scene
from satpy.tests.utils import make_dsq
from satpy.writers.cf_writer import _get_backend_versions

try:
    from pyproj import CRS
except ImportError:
    CRS = None

# NOTE:
# The following fixtures are not defined in this file, but are used and injected by Pytest:
# - tmp_path
# - request


class TempFile:
    """A temporary filename class."""

    def __init__(self, suffix=".nc"):
        """Initialize."""
        self.filename = None
        self.suffix = suffix

    def __enter__(self):
        """Enter."""
        self.handle, self.filename = tempfile.mkstemp(suffix=self.suffix)
        os.close(self.handle)
        return self.filename

    def __exit__(self, *args):
        """Exit."""
        os.remove(self.filename)


class TestCFWriter:
    """Test case for CF writer."""

    def test_init(self):
        """Test initializing the CFWriter class."""
        from satpy.writers.cf_writer import CFWriter
        from satpy.writers.core.config import configs_for_writer

        CFWriter(config_files=list(configs_for_writer("cf"))[0])

    def test_save_array(self):
        """Test saving an array to netcdf/cf."""
        scn = Scene()
        start_time = dt.datetime(2018, 5, 30, 10, 0)
        end_time = dt.datetime(2018, 5, 30, 10, 15)
        scn["test-array"] = xr.DataArray(da.from_array([1.0, 2.0, 3.0]),
                                         attrs=dict(start_time=start_time,
                                                    end_time=end_time,
                                                    prerequisites=[make_dsq(name="hej")]))
        with TempFile() as filename:
            scn.save_datasets(filename=filename, writer="cf")
            with xr.open_dataset(filename) as f:
                np.testing.assert_array_equal(f["test-array"][:], [1, 2, 3])
                expected_prereq = ("DataQuery(name='hej')")
                assert f["test-array"].attrs["prerequisites"] == expected_prereq

    def test_save_array_coords(self):
        """Test saving array with coordinates."""
        scn = Scene()
        start_time = dt.datetime(2018, 5, 30, 10, 0)
        end_time = dt.datetime(2018, 5, 30, 10, 15)
        coords = {
            "x": np.arange(3),
            "y": np.arange(1),
        }
        if CRS is not None:
            proj_str = ("+proj=geos +lon_0=-95.0 +h=35786023.0 "
                        "+a=6378137.0 +b=6356752.31414 +sweep=x "
                        "+units=m +no_defs")
            coords["crs"] = CRS.from_string(proj_str)
        scn["test-array"] = xr.DataArray(da.from_array([[1.0, 2.0, 3.0]]),
                                         dims=("y", "x"),
                                         coords=coords,
                                         attrs=dict(start_time=start_time,
                                                    end_time=end_time,
                                                    prerequisites=[make_dsq(name="hej")]))
        with TempFile() as filename:
            scn.save_datasets(filename=filename, writer="cf")
            with xr.open_dataset(filename) as f:
                np.testing.assert_array_equal(f["test-array"][:], [[1, 2, 3]])
                np.testing.assert_array_equal(f["x"][:], [0, 1, 2])
                np.testing.assert_array_equal(f["y"][:], [0])
                assert "crs" not in f
                assert "_FillValue" not in f["x"].attrs
                assert "_FillValue" not in f["y"].attrs
                expected_prereq = ("DataQuery(name='hej')")
                assert f["test-array"].attrs["prerequisites"] == expected_prereq

    def test_save_dataset_a_digit(self):
        """Test saving an array to netcdf/cf where dataset name starting with a digit."""
        scn = Scene()
        scn["1"] = xr.DataArray(da.from_array([1.0, 2.0, 3.0]))
        with TempFile() as filename:
            scn.save_datasets(filename=filename, writer="cf")
            with xr.open_dataset(filename) as f:
                np.testing.assert_array_equal(f["CHANNEL_1"][:], [1, 2, 3])

    def test_save_dataset_a_digit_prefix(self):
        """Test saving an array to netcdf/cf where dataset name starting with a digit with prefix."""
        scn = Scene()
        scn["1"] = xr.DataArray(da.from_array([1.0, 2.0, 3.0]))
        with TempFile() as filename:
            scn.save_datasets(filename=filename, writer="cf", numeric_name_prefix="TEST")
            with xr.open_dataset(filename) as f:
                np.testing.assert_array_equal(f["TEST1"][:], [1, 2, 3])

    def test_save_dataset_a_digit_prefix_include_attr(self):
        """Test saving an array to netcdf/cf where dataset name starting with a digit with prefix include orig name."""
        scn = Scene()
        scn["1"] = xr.DataArray(da.from_array([1.0, 2.0, 3.0]))
        with TempFile() as filename:
            scn.save_datasets(filename=filename, writer="cf", include_orig_name=True, numeric_name_prefix="TEST")
            with xr.open_dataset(filename) as f:
                np.testing.assert_array_equal(f["TEST1"][:], [1, 2, 3])
                assert f["TEST1"].attrs["original_name"] == "1"

    def test_save_dataset_a_digit_no_prefix_include_attr(self):
        """Test saving an array to netcdf/cf dataset name starting with a digit with no prefix include orig name."""
        scn = Scene()
        scn["1"] = xr.DataArray(da.from_array([1.0, 2.0, 3.0]))
        with TempFile() as filename:
            with pytest.warns(UserWarning, match=r"Invalid NetCDF dataset name"):
                scn.save_datasets(filename=filename, writer="cf", include_orig_name=True, numeric_name_prefix="")
            with xr.open_dataset(filename) as f:
                np.testing.assert_array_equal(f["1"][:], [1, 2, 3])
                assert "original_name" not in f["1"].attrs

    def test_ancillary_variables(self):
        """Test ancillary_variables cited each other."""
        from satpy.tests.utils import make_dataid
        scn = Scene()
        start_time = dt.datetime(2018, 5, 30, 10, 0)
        end_time = dt.datetime(2018, 5, 30, 10, 15)
        data_arr = xr.DataArray(
            da.from_array([1.0, 2.0, 3.0]),
            attrs=dict(start_time=start_time,
                       end_time=end_time,
                       prerequisites=[make_dataid(name="hej")]))
        scn["test-array-1"] = data_arr
        scn["test-array-2"] = data_arr.copy()
        scn["test-array-1"].attrs["ancillary_variables"] = [scn["test-array-2"]]
        scn["test-array-2"].attrs["ancillary_variables"] = [scn["test-array-1"]]
        with TempFile() as filename:
            scn.save_datasets(filename=filename, writer="cf")
            with xr.open_dataset(filename) as f:
                assert f["test-array-1"].attrs["ancillary_variables"] == "test-array-2"
                assert f["test-array-2"].attrs["ancillary_variables"] == "test-array-1"

    def test_groups(self):
        """Test creating a file with groups."""
        tstart = dt.datetime(2019, 4, 1, 12, 0)
        tend = dt.datetime(2019, 4, 1, 12, 15)

        data_visir = da.from_array([[1.0, 2.0], [3.0, 4.0]])
        y_visir = [1, 2]
        x_visir = [1, 2]
        time_vis006 = [1, 2]
        time_ir_108 = [3, 4]

        data_hrv = da.from_array([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 9.0]])
        y_hrv = [1, 2, 3]
        x_hrv = [1, 2, 3]
        time_hrv = [1, 2, 3]

        scn = Scene()
        scn["VIS006"] = xr.DataArray(data_visir,
                                     dims=("y", "x"),
                                     coords={"y": y_visir, "x": x_visir, "acq_time": ("y", time_vis006)},
                                     attrs={"name": "VIS006", "start_time": tstart, "end_time": tend})
        scn["IR_108"] = xr.DataArray(data_visir,
                                     dims=("y", "x"),
                                     coords={"y": y_visir, "x": x_visir, "acq_time": ("y", time_ir_108)},
                                     attrs={"name": "IR_108", "start_time": tstart, "end_time": tend})
        scn["HRV"] = xr.DataArray(data_hrv,
                                  dims=("y", "x"),
                                  coords={"y": y_hrv, "x": x_hrv, "acq_time": ("y", time_hrv)},
                                  attrs={"name": "HRV", "start_time": tstart, "end_time": tend})

        with TempFile() as filename:
            with pytest.warns(UserWarning, match=r"Cannot pretty-format"):
                scn.save_datasets(filename=filename, writer="cf",
                                  groups={"visir": ["IR_108", "VIS006"], "hrv": ["HRV"]},
                                  pretty=True)

            nc_root = xr.open_dataset(filename)
            assert "history" in nc_root.attrs
            assert set(nc_root.variables.keys()) == set()

            nc_visir = xr.open_dataset(filename, group="visir")
            nc_hrv = xr.open_dataset(filename, group="hrv")
            assert set(nc_visir.variables.keys()) == {"VIS006", "IR_108",
                                                      "y", "x", "VIS006_acq_time", "IR_108_acq_time"}
            assert set(nc_hrv.variables.keys()) == {"HRV", "y", "x", "acq_time"}
            for tst, ref in zip([nc_visir["VIS006"], nc_visir["IR_108"], nc_hrv["HRV"]],
                                [scn["VIS006"], scn["IR_108"], scn["HRV"]]):
                np.testing.assert_array_equal(tst.data, ref.data)
            nc_root.close()
            nc_visir.close()
            nc_hrv.close()

        # Different projection coordinates in one group are not supported
        with TempFile() as filename:
            with pytest.raises(ValueError, match="Datasets .* must have identical projection coordinates..*"):
                scn.save_datasets(datasets=["VIS006", "HRV"], filename=filename, writer="cf")

    def test_single_time_value(self):
        """Test setting a single time value."""
        scn = Scene()
        start_time = dt.datetime(2018, 5, 30, 10, 0)
        end_time = dt.datetime(2018, 5, 30, 10, 15)
        test_array = da.from_array(np.array([[1.0, 2.0], [3.0, 4.0]]))
        scn["test-array"] = xr.DataArray(test_array,
                                         dims=["x", "y"],
                                         coords={"time": np.datetime64("2018-05-30T10:05:00", "ns")},
                                         attrs=dict(start_time=start_time,
                                                    end_time=end_time))
        with TempFile() as filename:
            scn.save_datasets(filename=filename, writer="cf", encoding={"time": {"units": "seconds since 2018-01-01"}})
            with xr.open_dataset(filename, decode_cf=True) as f:
                np.testing.assert_array_equal(f["time"], scn["test-array"]["time"])
                bounds_exp = np.array([[start_time, end_time]], dtype="datetime64[m]")
                np.testing.assert_array_equal(f["time_bnds"], bounds_exp)

    def test_time_coordinate_on_a_swath(self):
        """Test that time dimension is not added on swath data with time already as a coordinate."""
        scn = Scene()
        test_array = np.array([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0], [7.0, 8.0]])
        times = np.array(["2018-05-30T10:05:00", "2018-05-30T10:05:01",
                          "2018-05-30T10:05:02", "2018-05-30T10:05:03"], dtype="datetime64[ns]")
        scn["test-array"] = xr.DataArray(test_array,
                                         dims=["y", "x"],
                                         coords={"time": ("y", times)},
                                         attrs=dict(start_time=times[0], end_time=times[-1]))
        with TempFile() as filename:
            scn.save_datasets(filename=filename, writer="cf", pretty=True,
                              encoding={"time": {"units": "seconds since 2018-01-01"}})
            with xr.open_dataset(filename, decode_cf=True) as f:
                np.testing.assert_array_equal(f["time"], scn["test-array"]["time"])

    def test_bounds(self):
        """Test setting time bounds."""
        scn = Scene()
        start_time = dt.datetime(2018, 5, 30, 10, 0)
        end_time = dt.datetime(2018, 5, 30, 10, 15)
        test_array = np.array([[1.0, 2.0], [3.0, 4.0]]).reshape(2, 2, 1)
        scn["test-array"] = xr.DataArray(test_array,
                                         dims=["x", "y", "time"],
                                         coords={"time": [np.datetime64("2018-05-30T10:05:00", "ns")]},
                                         attrs=dict(start_time=start_time,
                                                    end_time=end_time))
        with TempFile() as filename:
            with warnings.catch_warnings():
                # The purpose is to use the default time encoding, silence the warning
                warnings.filterwarnings("ignore", category=UserWarning,
                                        message=r"Times can't be serialized faithfully to int64 with requested units")
                scn.save_datasets(filename=filename, writer="cf")
            # Check decoded time coordinates & bounds
            with xr.open_dataset(filename, decode_cf=True) as f:
                bounds_exp = np.array([[start_time, end_time]], dtype="datetime64[m]")
                np.testing.assert_array_equal(f["time_bnds"], bounds_exp)
                assert f["time"].attrs["bounds"] == "time_bnds"

            # Check raw time coordinates & bounds
            with xr.open_dataset(filename, decode_cf=False) as f:
                np.testing.assert_almost_equal(f["time_bnds"], [[-0.0034722, 0.0069444]])

        # User-specified time encoding should have preference
        with TempFile() as filename:
            time_units = "seconds since 2018-01-01"
            scn.save_datasets(filename=filename, encoding={"time": {"units": time_units}},
                              writer="cf")
            with xr.open_dataset(filename, decode_cf=False) as f:
                np.testing.assert_array_equal(f["time_bnds"], [[12909600, 12910500]])

    def test_bounds_minimum(self):
        """Test minimum bounds."""
        scn = Scene()
        start_timeA = dt.datetime(2018, 5, 30, 10, 0)  # expected to be used
        end_timeA = dt.datetime(2018, 5, 30, 10, 15)
        start_timeB = dt.datetime(2018, 5, 30, 10, 3)
        end_timeB = dt.datetime(2018, 5, 30, 10, 20)  # expected to be used
        test_arrayA = np.array([[1.0, 2.0], [3.0, 4.0]]).reshape(2, 2, 1)
        test_arrayB = np.array([[1.0, 2.0], [3.0, 5.0]]).reshape(2, 2, 1)
        scn["test-arrayA"] = xr.DataArray(test_arrayA,
                                          dims=["x", "y", "time"],
                                          coords={"time": [np.datetime64("2018-05-30T10:05:00", "ns")]},
                                          attrs=dict(start_time=start_timeA,
                                                     end_time=end_timeA))
        scn["test-arrayB"] = xr.DataArray(test_arrayB,
                                          dims=["x", "y", "time"],
                                          coords={"time": [np.datetime64("2018-05-30T10:05:00", "ns")]},
                                          attrs=dict(start_time=start_timeB,
                                                     end_time=end_timeB))
        with TempFile() as filename:
            scn.save_datasets(filename=filename, writer="cf",
                              encoding={"time": {"units": "seconds since 2018-01-01"}})
            with xr.open_dataset(filename, decode_cf=True) as f:
                bounds_exp = np.array([[start_timeA, end_timeB]], dtype="datetime64[m]")
                np.testing.assert_array_equal(f["time_bnds"], bounds_exp)

    def test_bounds_missing_time_info(self):
        """Test time bounds generation in case of missing time."""
        scn = Scene()
        start_timeA = dt.datetime(2018, 5, 30, 10, 0)
        end_timeA = dt.datetime(2018, 5, 30, 10, 15)
        test_arrayA = np.array([[1.0, 2.0], [3.0, 4.0]]).reshape(2, 2, 1)
        test_arrayB = np.array([[1.0, 2.0], [3.0, 5.0]]).reshape(2, 2, 1)
        scn["test-arrayA"] = xr.DataArray(test_arrayA,
                                          dims=["x", "y", "time"],
                                          coords={"time": [np.datetime64("2018-05-30T10:05:00", "ns")]},
                                          attrs=dict(start_time=start_timeA,
                                                     end_time=end_timeA))
        scn["test-arrayB"] = xr.DataArray(test_arrayB,
                                          dims=["x", "y", "time"],
                                          coords={"time": [np.datetime64("2018-05-30T10:05:00", "ns")]})
        with TempFile() as filename:
            scn.save_datasets(filename=filename, writer="cf",
                              encoding={"time": {"units": "seconds since 2018-01-01"}})
            with xr.open_dataset(filename, decode_cf=True) as f:
                bounds_exp = np.array([[start_timeA, end_timeA]], dtype="datetime64[m]")
                np.testing.assert_array_equal(f["time_bnds"], bounds_exp)

    def test_unlimited_dims_kwarg(self):
        """Test specification of unlimited dimensions."""
        scn = Scene()
        start_time = dt.datetime(2018, 5, 30, 10, 0)
        end_time = dt.datetime(2018, 5, 30, 10, 15)
        test_array = np.array([[1.0, 2.0], [3.0, 4.0]])
        scn["test-array"] = xr.DataArray(test_array,
                                         dims=["x", "y"],
                                         coords={"time": np.datetime64("2018-05-30T10:05:00", "ns")},
                                         attrs=dict(start_time=start_time,
                                                    end_time=end_time))
        with TempFile() as filename:
            scn.save_datasets(filename=filename, writer="cf", unlimited_dims=["time"],
                              encoding={"time": {"units": "seconds since 2018-01-01"}})
            with xr.open_dataset(filename) as f:
                assert set(f.encoding["unlimited_dims"]) == {"time"}

    def test_header_attrs(self):
        """Check global attributes are set."""
        scn = Scene()
        start_time = dt.datetime(2018, 5, 30, 10, 0)
        end_time = dt.datetime(2018, 5, 30, 10, 15)
        scn["test-array"] = xr.DataArray([1.0, 2.0, 3.0],
                                         attrs=dict(start_time=start_time,
                                                    end_time=end_time))
        with TempFile() as filename:
            header_attrs = {"sensor": "SEVIRI",
                            "orbit": 99999,
                            "none": None,
                            "list": [1, 2, 3],
                            "set": {1, 2, 3},
                            "dict": {"a": 1, "b": 2},
                            "nested": {"outer": {"inner1": 1, "inner2": 2}},
                            "bool": True,
                            "bool_": np.bool_(True)}
            scn.save_datasets(filename=filename,
                              header_attrs=header_attrs,
                              flatten_attrs=True,
                              writer="cf")
            with xr.open_dataset(filename) as f:
                assert "history" in f.attrs
                assert f.attrs["sensor"] == "SEVIRI"
                assert f.attrs["orbit"] == 99999
                np.testing.assert_array_equal(f.attrs["list"], [1, 2, 3])
                assert f.attrs["set"] == "{1, 2, 3}"
                assert f.attrs["dict_a"] == 1
                assert f.attrs["dict_b"] == 2
                assert f.attrs["nested_outer_inner1"] == 1
                assert f.attrs["nested_outer_inner2"] == 2
                assert f.attrs["bool"] == "true"
                assert f.attrs["bool_"] == "true"
                assert "none" not in f.attrs.keys()

    def test_load_module_with_old_pyproj(self):
        """Test that cf_writer can still be loaded with pyproj 1.9.6."""
        import importlib
        import sys

        import pyproj  # noqa 401
        old_version = sys.modules["pyproj"].__version__
        sys.modules["pyproj"].__version__ = "1.9.6"
        try:
            importlib.reload(sys.modules["satpy.writers.cf_writer"])
        finally:
            # Tear down
            sys.modules["pyproj"].__version__ = old_version
            importlib.reload(sys.modules["satpy.writers.cf_writer"])

    def test_global_attr_default_history_and_Conventions(self):
        """Test saving global attributes history and Conventions."""
        scn = Scene()
        start_time = dt.datetime(2018, 5, 30, 10, 0)
        end_time = dt.datetime(2018, 5, 30, 10, 15)
        scn["test-array"] = xr.DataArray([[1.0, 2.0, 3.0]],
                                         dims=("y", "x"),
                                         attrs=dict(start_time=start_time,
                                                    end_time=end_time,
                                                    prerequisites=[make_dsq(name="hej")]))
        with TempFile() as filename:
            scn.save_datasets(filename=filename, writer="cf")
            with xr.open_dataset(filename) as f:
                assert f.attrs["Conventions"] == "CF-1.7"
                assert "Created by pytroll/satpy on" in f.attrs["history"]

    def test_global_attr_history_and_Conventions(self):
        """Test saving global attributes history and Conventions."""
        scn = Scene()
        start_time = dt.datetime(2018, 5, 30, 10, 0)
        end_time = dt.datetime(2018, 5, 30, 10, 15)
        scn["test-array"] = xr.DataArray([[1.0, 2.0, 3.0]],
                                         dims=("y", "x"),
                                         attrs=dict(start_time=start_time,
                                                    end_time=end_time,
                                                    prerequisites=[make_dsq(name="hej")]))
        header_attrs = {}
        header_attrs["history"] = ("TEST add history",)
        header_attrs["Conventions"] = "CF-1.7, ACDD-1.3"
        with TempFile() as filename:
            scn.save_datasets(filename=filename, writer="cf", header_attrs=header_attrs)
            with xr.open_dataset(filename) as f:
                assert f.attrs["Conventions"] == "CF-1.7, ACDD-1.3"
                assert "TEST add history\n" in f.attrs["history"]
                assert "Created by pytroll/satpy on" in f.attrs["history"]


class TestNetcdfEncodingKwargs:
    """Test netCDF compression encodings."""

    @pytest.fixture
    def scene(self):
        """Create a fake scene."""
        scn = Scene()
        attrs = {
            "start_time": dt.datetime(2018, 5, 30, 10, 0),
            "end_time": dt.datetime(2018, 5, 30, 10, 15)
        }
        scn["test-array"] = xr.DataArray([1.0, 2.0, 3.0], attrs=attrs)
        return scn

    @pytest.fixture(params=[True, False])
    def compression_on(self, request):
        """Get compression options."""
        return request.param

    @pytest.fixture
    def encoding(self, compression_on):
        """Get encoding."""
        enc = {
            "test-array": {
                "dtype": "int8",
                "scale_factor": 0.1,
                "add_offset": 0.0,
                "_FillValue": 3,
            }
        }
        if compression_on:
            comp_params = _get_compression_params(complevel=7)
            enc["test-array"].update(comp_params)
        return enc

    @pytest.fixture
    def filename(self, tmp_path):
        """Get output filename."""
        return str(tmp_path / "test.nc")

    @pytest.fixture
    def complevel_exp(self, compression_on):
        """Get expected compression level."""
        if compression_on:
            return 7
        return 0

    @pytest.fixture
    def expected(self, complevel_exp):
        """Get expectated file contents."""
        return {
            "data": [10, 20, 30],
            "scale_factor": 0.1,
            "fill_value": 3,
            "dtype": np.int8,
            "complevel": complevel_exp
        }

    def test_encoding_kwarg(self, scene, encoding, filename, expected):
        """Test 'encoding' keyword argument."""
        scene.save_datasets(filename=filename, encoding=encoding, writer="cf")
        self._assert_encoding_as_expected(filename, expected)

    def _assert_encoding_as_expected(self, filename, expected):
        with xr.open_dataset(filename, mask_and_scale=False) as f:
            np.testing.assert_array_equal(f["test-array"][:], expected["data"])
            assert f["test-array"].attrs["scale_factor"] == expected["scale_factor"]
            assert f["test-array"].attrs["_FillValue"] == expected["fill_value"]
            assert f["test-array"].dtype == expected["dtype"]
            if "complevel" in f["test-array"].encoding:
                # netcdf4 engine and h5netcdf engine (when not 0)
                assert f["test-array"].encoding["complevel"] == expected["complevel"]
            else:
                assert expected["complevel"] == 0, "No compression option found when compression was non-zero"


    @pytest.mark.parametrize(
        "versions",
        [
            {"netCDF4": "1.5.0", "libnetcdf": "4.9.1-development",
             "warning_msg": "Backend version mismatch"},
            {"netCDF4": "1.6.0", "libnetcdf": "invalid-version",
             "warning_msg": "Unable to parse netcdf-c"},
        ]
    )
    def test_warning_if_backends_dont_match(self, scene, filename, monkeypatch, versions):
        """Test warning if backends don't match."""
        import netCDF4
        invalid_catch = nullcontext()
        if "invalid" in versions["libnetcdf"]:
            invalid_catch = pytest.warns(UserWarning, match="invalid-version")
        with monkeypatch.context() as m:
            m.setattr(netCDF4, "__version__", versions["netCDF4"])
            m.setattr(netCDF4, "__netcdf4libversion__", versions["libnetcdf"])
            with pytest.warns(UserWarning, match="Backend version mismatch"), \
                    invalid_catch:
                scene.save_datasets(filename=filename, writer="cf")

    def test_no_warning_if_backends_match(self, scene, filename, monkeypatch):
        """Make sure no warning is issued if backends match."""
        import netCDF4
        with monkeypatch.context() as m:
            m.setattr(netCDF4, "__version__", "1.6.0")
            m.setattr(netCDF4, "__netcdf4libversion__", "4.9.0")
            m.setattr(xr, "__version__", "2022.12.0")
            with warnings.catch_warnings():
                scene.save_datasets(filename=filename, writer="cf")
                warnings.simplefilter("error")


class TestEncodingAttribute(TestNetcdfEncodingKwargs):
    """Test CF writer with 'encoding' dataset attribute."""

    @pytest.fixture
    def scene_with_encoding(self, scene, encoding):
        """Create scene with a dataset providing the 'encoding' attribute."""
        scene["test-array"].encoding = encoding["test-array"]
        return scene

    def test_encoding_attribute(self, scene_with_encoding, filename, expected):
        """Test 'encoding' dataset attribute."""
        scene_with_encoding.save_datasets(filename=filename, writer="cf")
        self._assert_encoding_as_expected(filename, expected)


def _get_compression_params(complevel):
    params = {"complevel": complevel}
    if _should_use_compression_keyword():
        params["compression"] = "zlib"
    else:
        params["zlib"] = True
    return params


def _should_use_compression_keyword():
    # xarray currently ignores the "compression" keyword, see
    # https://github.com/pydata/xarray/issues/7388. There's already an open
    # PR, so we assume that this will be fixed in the next minor release
    # (current release is 2023.02). If not, tests will fail and remind us.
    versions = _get_backend_versions()
    return (
        versions["libnetcdf"] >= Version("4.9.0") and
        versions["xarray"] >= Version("2024.1")
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2016-2018 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Module for testing the satpy.readers.goci2_l2_nc module."""

import datetime as dt

import numpy as np
import pytest
import xarray as xr
from pytest_lazy_fixtures import lf as lazy_fixture

from satpy import Scene
from satpy.tests.utils import RANDOM_GEN

# NOTE:
# The following fixtures are not defined in this file, but are used and injected by Pytest:
# - tmp_path_factory


start_time = dt.datetime(2024, 2, 14, 2, 32, 27)
end_time = dt.datetime(2024, 2, 14, 2, 33, 31)

global_attrs = {
    "observation_start_time": start_time.strftime("%Y%m%d_%H%M%S"),
    "observation_end_time": end_time.strftime("%Y%m%d_%H%M%S"),
    "instrument": "GOCI-II",
    "platform": "GK-2B",
}

badarea_attrs = global_attrs.copy()
badarea_attrs["cdm_data_type"] = "bad_area"


def _create_lonlat():
    """Create a fake navigation dataset with lon/lat."""
    lon, lat = np.meshgrid(np.linspace(120, 130, 10), np.linspace(30, 40, 10))
    lon = xr.DataArray(
        lon,
        dims=("number_of_lines", "pixels_per_line"),
        attrs={"standard_name": "longitude", "units": "degrees_east"},
    )
    lat = xr.DataArray(
        lat,
        dims=("number_of_lines", "pixels_per_line"),
        attrs={"standard_name": "latitude", "units": "degrees_north"},
    )
    ds = xr.Dataset()
    ds["longitude"] = lon
    ds["latitude"] = lat
    return ds


def _create_bad_lon_lat():
    """Create a fake navigation dataset with lon/lat base name missing."""
    lon, lat = np.meshgrid(np.linspace(120, 130, 10), np.linspace(30, 40, 10))
    ds = xr.Dataset(
        {
            "longitude": (["number_of_lines", "pixels_per_line"], lon),
            "latitude": (["number_of_lines", "pixels_per_line"], lat),
        }
    )
    return ds


@pytest.fixture(scope="session")
def ac_file(tmp_path_factory):
    """Create a fake atmospheric correction product."""
    data = RANDOM_GEN.random((10, 10))
    RhoC = xr.Dataset(
        {"RhoC_555": (["number_of_lines", "pixels_per_line"], data)},
        coords={"number_of_lines": np.arange(10), "pixels_per_line": np.arange(10)},
    )
    Rrs = xr.Dataset(
        {"Rrs_555": (["number_of_lines", "pixels_per_line"], data)},
        coords={"number_of_lines": np.arange(10), "pixels_per_line": np.arange(10)},
    )
    navigation = _create_lonlat()
    ds = xr.Dataset(attrs=global_attrs)
    fname = (
        f'{tmp_path_factory.mktemp("data")}/GK2B_GOCI2_L2_20240214_021530_LA_S010_AC.nc'
    )
    ds.to_netcdf(fname)
    navigation.to_netcdf(fname, group="navigation_data", mode="a")
    RhoC.to_netcdf(fname, group="geophysical_data/RhoC", mode="a")
    Rrs.to_netcdf(fname, group="geophysical_data/Rrs", mode="a")
    return fname


@pytest.fixture(scope="module")
def iop_file(tmp_path_factory):
    """Create a fake IOP product."""
    data = RANDOM_GEN.random((10, 10))
    a = xr.Dataset(
        {"a_total_555": (["number_of_lines", "pixels_per_line"], data)},
        coords={"number_of_lines": np.arange(10), "pixels_per_line": np.arange(10)},
    )
    bb = xr.Dataset(
        {"bb_total_555": (["number_of_lines", "pixels_per_line"], data)},
        coords={"number_of_lines": np.arange(10), "pixels_per_line": np.arange(10)},
    )
    navigation = _create_lonlat()
    ds = xr.Dataset(attrs=global_attrs)
    fname = f'{tmp_path_factory.mktemp("data")}/GK2B_GOCI2_L2_20240214_021530_LA_S010_IOP.nc'
    ds.to_netcdf(fname)
    navigation.to_netcdf(fname, group="navigation_data", mode="a")
    a.to_netcdf(fname, group="geophysical_data/a_total", mode="a")
    bb.to_netcdf(fname, group="geophysical_data/bb_total", mode="a")
    return fname


@pytest.fixture(scope="module")
def generic_file(tmp_path_factory):
    """Create a fake ouput product like Chl, Zsd etc."""
    data = RANDOM_GEN.random((10, 10))
    geophysical_data = xr.Dataset(
        {"Chl": (["number_of_lines", "pixels_per_line"], data)},
        coords={"number_of_lines": np.arange(10), "pixels_per_line": np.arange(10)},
    )
    navigation = _create_lonlat()
    ds = xr.Dataset(attrs=global_attrs)
    fname = f'{tmp_path_factory.mktemp("data")}/GK2B_GOCI2_L2_20240214_021530_LA_S010_Chl.nc'
    ds.to_netcdf(fname)
    navigation.to_netcdf(fname, group="navigation_data", mode="a")
    geophysical_data.to_netcdf(fname, group="geophysical_data", mode="a")
    return fname


@pytest.fixture(scope="module")
def generic_bad_file(tmp_path_factory):
    """Create a PP product with lon/lat base name missing."""
    data = RANDOM_GEN.random((10, 10))
    geophysical_data = xr.Dataset(
        {"PP": (["number_of_lines", "pixels_per_line"], data)},
        coords={"number_of_lines": np.arange(10), "pixels_per_line": np.arange(10)},
    )
    navigation = _create_bad_lon_lat()
    ds = xr.Dataset(attrs=global_attrs)
    fname = (
        f'{tmp_path_factory.mktemp("data")}/GK2B_GOCI2_L2_20240214_021530_LA_S010_PP.nc'
    )
    ds.to_netcdf(fname)
    navigation.to_netcdf(fname, group="navigation_data", mode="a")
    geophysical_data.to_netcdf(fname, group="geophysical_data", mode="a")
    return fname


class TestGOCI2Reader:
    """Test the GOCI-II L2 netcdf file reader."""

    @pytest.mark.parametrize(
        "test_files",
        [
            lazy_fixture("ac_file"),
            lazy_fixture("iop_file"),
            lazy_fixture("generic_file"),
            lazy_fixture("generic_bad_file"),
        ],
    )
    def test_scene_available_datasets(self, test_files):
        """Test that datasets are available."""
        scene = Scene(filenames=[test_files], reader="goci2_l2_nc")
        available_datasets = scene.all_dataset_names()
        assert len(available_datasets) > 0
        assert "longitude" in available_datasets
        assert "latitude" in available_datasets

    @pytest.mark.parametrize(
        "test_files",
        [
            lazy_fixture("ac_file"),
            lazy_fixture("iop_file"),
            lazy_fixture("generic_file"),
            lazy_fixture("generic_bad_file"),
        ],
    )
    def test_start_end_time(self, test_files):
        """Test dataset start_time and end_time."""
        scene = Scene(filenames=[test_files], reader="goci2_l2_nc")
        assert scene.start_time == start_time
        assert scene.end_time == end_time

    @pytest.mark.parametrize(
        ("test_files", "datasets"),
        [
            (lazy_fixture("ac_file"), ["RhoC_555", "Rrs_555"]),
            (lazy_fixture("iop_file"), ["a_total_555", "bb_total_555"]),
            (lazy_fixture("generic_file"), ["Chl"]),
            (lazy_fixture("generic_bad_file"), ["PP"]),
        ],
    )
    def test_load_dataset(self, test_files, datasets):
        """Test dataset loading."""
        scene = Scene(filenames=[test_files], reader="goci2_l2_nc")
        scene.load(datasets)
        for dataset in datasets:
            data_arr = scene[dataset]
            assert data_arr.dims == ("y", "x")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Tests for the simple image writer."""
import unittest


class TestPillowWriter(unittest.TestCase):
    """Test Pillow/PIL writer."""

    def setUp(self):
        """Create temporary directory to save files to."""
        import tempfile
        self.base_dir = tempfile.mkdtemp()

    def tearDown(self):
        """Remove the temporary directory created for a test."""
        try:
            import shutil
            shutil.rmtree(self.base_dir, ignore_errors=True)
        except OSError:
            pass

    @staticmethod
    def _get_test_datasets():
        """Create DataArray for testing."""
        import datetime as dt

        import dask.array as da
        import xarray as xr
        ds1 = xr.DataArray(
            da.arange(100 * 200).reshape((100, 200)).rechunk(50),
            dims=("y", "x"),
            attrs={"name": "test",
                   "start_time": dt.datetime.now(dt.timezone.utc)}
        )
        return [ds1]

    def test_init(self):
        """Test creating the default writer."""
        from satpy.writers.simple_image import PillowWriter
        PillowWriter()

    def test_simple_write(self):
        """Test writing datasets with default behavior."""
        from satpy.writers.simple_image import PillowWriter
        datasets = self._get_test_datasets()
        w = PillowWriter(base_dir=self.base_dir)
        w.save_datasets(datasets)

    def test_simple_delayed_write(self):
        """Test writing datasets with delayed computation."""
        import dask.array as da
        from dask.delayed import Delayed

        from satpy.writers.core.compute import compute_writer_results
        from satpy.writers.simple_image import PillowWriter
        datasets = self._get_test_datasets()
        w = PillowWriter(base_dir=self.base_dir)
        res = w.save_datasets(datasets, compute=False)
        for r__ in res:
            # trollimage 1.27.0+ returns Arrays
            # trollimage <1.27.0 returns Delayed objects
            assert isinstance(r__, (Delayed, da.Array))
            r__.compute()
        compute_writer_results([res])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2020 Satpy developers
#
# satpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# satpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with satpy.  If not, see <http://www.gnu.org/licenses/>.
"""The vii_l1b_nc reader tests package.

This version tests the readers for VII test data V2 as per PFS V4A.

"""


import datetime
import os
import unittest
import uuid

import dask.array as da
import numpy as np
import pytest
import xarray as xr
from netCDF4 import Dataset

from satpy.readers.core.vii import MEAN_EARTH_RADIUS
from satpy.readers.vii_l1b_nc import ViiL1bNCFileHandler

TEST_FILE = "test_file_vii_l1b_nc.nc"


class TestViiL1bNCFileHandler(unittest.TestCase):
    """Test the ViiL1bNCFileHandler reader."""

    def setUp(self):
        """Set up the test."""
        # Easiest way to test the reader is to create a test netCDF file on the fly
        # uses a UUID to avoid permission conflicts during execution of tests in parallel
        self.test_file_name = TEST_FILE + str(uuid.uuid1()) + ".nc"

        with Dataset(self.test_file_name, "w") as nc:
            # Create data group
            g1 = nc.createGroup("data")

            # Add dimensions to data group
            g1.createDimension("num_chan_solar", 11)
            g1.createDimension("num_chan_thermal", 9)
            g1.createDimension("num_pixels", 72)
            g1.createDimension("num_lines", 600)

            # Create calibration_data group
            g1_1 = g1.createGroup("calibration_data")

            # Add variables to data/calibration_data group
            bt_a = g1_1.createVariable("bt_conversion_a", np.float32, dimensions=("num_chan_thermal",))
            bt_a[:] = np.arange(9)
            bt_b = g1_1.createVariable("bt_conversion_b", np.float32, dimensions=("num_chan_thermal",))
            bt_b[:] = np.arange(9)
            cw = g1_1.createVariable("channel_cw_thermal", np.float32, dimensions=("num_chan_thermal",))
            cw[:] = np.arange(9)
            isi = g1_1.createVariable("band_averaged_solar_irradiance", np.float32, dimensions=("num_chan_solar",))
            isi[:] = np.arange(11)

            # Create measurement_data group
            g1_2 = g1.createGroup("measurement_data")

            # Add dimensions to data/measurement_data group
            g1_2.createDimension("num_tie_points_act", 10)
            g1_2.createDimension("num_tie_points_alt", 100)

            # Add variables to data/measurement_data group
            sza = g1_2.createVariable("solar_zenith", np.float32,
                                      dimensions=("num_tie_points_alt", "num_tie_points_act"))
            sza[:] = 25.0
            delta_lat = g1_2.createVariable("delta_lat", np.float32, dimensions=("num_lines", "num_pixels"))
            delta_lat[:] = 1.0

        self.reader = ViiL1bNCFileHandler(
            filename=self.test_file_name,
            filename_info={
                "creation_time": datetime.datetime(year=2017, month=9, day=22,
                                                   hour=22, minute=40, second=10),
                "sensing_start_time": datetime.datetime(year=2017, month=9, day=20,
                                                        hour=12, minute=30, second=30),
                "sensing_end_time": datetime.datetime(year=2017, month=9, day=20,
                                                      hour=18, minute=30, second=50)
            },
            filetype_info={}
        )

    def tearDown(self):
        """Remove the previously created test file."""
        # Catch Windows PermissionError for removing the created test file.
        try:
            os.remove(self.test_file_name)
        except OSError:
            pass

    def test_calibration_functions(self):
        """Test the calibration functions."""
        radiance = np.array([[1.0, 2.0, 5.0], [7.0, 10.0, 20.0]])

        cw = 13.0
        a = 3.0
        b = 100.0
        bt = self.reader._calibrate_bt(radiance, cw, a, b)
        expected_bt = np.array([[675.04993213, 753.10301462, 894.93149648],
                                [963.20401882, 1048.95086402, 1270.95546218]])
        assert np.allclose(bt, expected_bt)

        isi = 2.0
        refl = self.reader._calibrate_refl(radiance, isi)
        expected_refl = np.array([[157.07963268, 314.15926536, 785.3981634],
                                  [1099.55742876, 1570.79632679, 3141.59265359]])
        assert np.allclose(refl, expected_refl)

    def test_functions(self):
        """Test the functions."""
        # Checks that the _perform_orthorectification function is correctly executed
        variable = xr.DataArray(
            dims=("num_lines", "num_pixels"),
            name="test_name",
            attrs={
                "key_1": "value_1",
                "key_2": "value_2"
            },
            data=da.from_array(np.ones((600, 72)))
        )

        orthorect_variable = self.reader._perform_orthorectification(variable, "data/measurement_data/delta_lat")
        expected_values = np.degrees(np.ones((600, 72)) / MEAN_EARTH_RADIUS) + np.ones((600, 72))
        assert np.allclose(orthorect_variable.values, expected_values)

        # Checks that the _perform_calibration function is correctly executed in all cases
        # radiance calibration: return value is simply a copy of the variable
        return_variable = self.reader._perform_calibration(variable, {"calibration": "radiance"})
        assert np.all(return_variable == variable)

        # invalid calibration: raises a ValueError
        with pytest.raises(ValueError, match="Unknown calibration invalid for dataset test"):
            self.reader._perform_calibration(variable,
                                             {"calibration": "invalid", "name": "test"})

        # brightness_temperature calibration: checks that the return value is correct
        calibrated_variable = self.reader._perform_calibration(variable,
                                                               {"calibration": "brightness_temperature",
                                                                "chan_thermal_index": 3})
        expected_values = np.full((600, 72), 1101.10413712)
        assert np.allclose(calibrated_variable.values, expected_values)

        # reflectance calibration: checks that the return value is correct
        calibrated_variable = self.reader._perform_calibration(variable,
                                                               {"calibration": "reflectance",
                                                                "wavelength": [0.658, 0.668, 0.678],
                                                                "chan_solar_index": 2})
        expected_values = np.full((600, 72), 157.07963268)
        assert np.allclose(calibrated_variable.values, expected_values)



# Copyright (c) 2022 Satpy developers
#
# satpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# satpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests on the LI L2 reader using the conventional mock constructed context."""

import datetime as dt
import os
from unittest import mock

import dask.array as da
import numpy as np
import pytest
import xarray as xr
from pyproj import Proj

from satpy._config import config_search_paths
from satpy.readers.core.li_nc import LINCFileHandler
from satpy.readers.core.yaml_reader import load_yaml_configs
from satpy.readers.li_l2_nc import LI_GRID_SHAPE, LIL2NCFileHandler
from satpy.tests.reader_tests._li_test_utils import (
    FakeLIFileHandlerBase,
    expected_product_dtype,
    extract_filetype_info,
    get_product_schema,
    products_dict,
)
from satpy.tests.utils import make_dataid


@pytest.fixture(name="filetype_infos")
def std_filetype_infos():
    """Return standard filetype info for LI L2."""
    cpaths = config_search_paths(os.path.join("readers", "li_l2_nc.yaml"))

    cfg = load_yaml_configs(cpaths[0])

    # get the li_l2 filetype:
    ftypes = cfg["file_types"]

    return ftypes


# Note: the helper class below has some missing abstract class implementation,
# but that is not critical to us, so ignoring them for now.
class TestLIL2():
    """Main test class for the LI L2 reader."""

    @pytest.fixture(autouse=True, scope="class")
    def fake_handler(self):
        """Wrap NetCDF4 FileHandler with our own fake handler."""
        patch_ctx = mock.patch.object(
            LINCFileHandler,
            "__bases__",
            (FakeLIFileHandlerBase,))

        with patch_ctx:
            patch_ctx.is_local = True
            yield patch_ctx

    def _test_dataset_single_variable(self, vname, desc, settings, handler):
        """Check the validity of a given variable."""
        dname = vname

        dims = settings.get("dimensions", {})

        var_path = settings.get("variable_path", "")

        # Compute shape from dimensions:
        if desc["shape"] == ():
            # scalar case, dim should have been added in the code by validate_array_dimensions
            shape = (1,)
        else:
            shape = tuple([dims[dim_name] for dim_name in desc["shape"]])

        dataset_info = {
            "name": dname,
            "variable_name": vname,
            "use_rescaling": False,
        }
        var_params = [dataset_info, desc, dname, handler, shape, var_path]
        self._test_dataset_variable(var_params)

    def _test_dataset_variables(self, settings, ds_desc, handler):
        """Check the loading of the non in sector variables."""
        assert "variables" in ds_desc
        all_vars = ds_desc["variables"]
        variables = settings.get("variables")
        for vname, desc in variables.items():
            # variable should be in list of dataset:
            assert vname in all_vars
            self._test_dataset_single_variable(vname, desc, settings, handler)

    def _test_dataset_single_sector_variable(self, names, desc, settings, handler):
        """Check the validity of a given sector variable."""
        sname, vname = names[0], names[1]

        dname = f"{vname}_{sname}_sector"

        dims = settings.get("dimensions", {})

        var_path = settings.get("variable_path", "")

        shape = tuple([dims[dim_name] for dim_name in desc["shape"]])

        dataset_info = {
            "name": dname,
            "variable_name": vname,
            "sector_name": sname,
            "use_rescaling": False,
        }
        var_params = [dataset_info, desc, vname, handler, shape, var_path]
        self._test_dataset_variable(var_params, sname=sname)

    def _test_dataset_variable(self, var_params, sname=""):
        """Test the validity of a given (sector) variable."""
        dataset_info, desc, dname, handler, shape, var_path = var_params
        product_type = handler.ds_desc["product_type"]
        res = self.get_variable_dataset(dataset_info, dname, handler)
        resd = self.get_variable_dataset(None,dataset_info["name"], handler)
        assert resd.dtype == expected_product_dtype[product_type][dname]
        assert res.shape == shape
        assert res.dims[0] == "y"
        assert isinstance(res.data,da.Array)
        # Should retrieve content with fullname key:
        full_name = self.create_fullname_key(desc, var_path, dname, sname=sname)
        # Note: 'content' is not recognized as a valid member of the class below
        # since it is silently injected in from our patching fake base netcdf4 file handler class.
        # But for now, we don't need to actually extend the class itself as this is only
        # needed for testing.
        assert np.all(res.values == handler.content[full_name])  # pylint: disable=no-member

    def get_variable_dataset(self, dataset_info, dname, handler):
        """Get the dataset of a given (sector) variable."""
        dataset_id = make_dataid(name=dname)
        res = handler.get_dataset(dataset_id, dataset_info)
        return res

    def create_fullname_key(self, desc, var_path, vname, sname=""):
        """Create full name key for sector/non-sector content retrieval."""
        vpath = desc.get("path", var_path)
        if vpath != "" and vpath[-1] != "/":
            vpath += "/"
        if sname != "":
            sname += "/"
        full_name = f"{vpath}{sname}{vname}"
        return full_name

    def _test_dataset_sector_variables(self, settings, ds_desc, handler):
        """Check the loading of the in sector variables."""
        sector_vars = settings.get("sector_variables")
        sectors = settings.get("sectors", ["north", "east", "south", "west"])

        assert "sector_variables" in ds_desc
        all_vars = ds_desc["sector_variables"]

        for sname in sectors:
            for vname, desc in sector_vars.items():
                # variable should be in list of dataset:
                assert vname in all_vars
                self._test_dataset_single_sector_variable([sname, vname], desc, settings, handler)

    def test_dataset_loading(self, filetype_infos):
        """Test loading of all datasets from all products."""
        # Iterate on all the available product types:
        for ptype, pinfo in products_dict.items():
            ftype = pinfo["ftype"]
            filename_info = {
                "start_time": "0000",
                "end_time": "1000",
                "mission_prefix": "MT",
                "spacecraft_id": "1"
            }
            handler = LIL2NCFileHandler("filename", filename_info, extract_filetype_info(filetype_infos, ftype),
                                       with_area_definition=False)
            ds_desc = handler.ds_desc

            # retrieve the schema that what used to generate the content for that product:
            settings = get_product_schema(ptype)

            # Now we check all the variables are available:
            if "variables" in settings:
                self._test_dataset_variables(settings, ds_desc, handler)

                # check the sector variables:
            if "sector_variables" in settings:
                self._test_dataset_sector_variables(settings, ds_desc, handler)

    def test_unregistered_dataset_loading(self, filetype_infos):
        """Test loading of an unregistered dataset."""
        filename_info = {
            "start_time": "0000",
            "end_time": "1000",
            "mission_prefix": "MT",
            "spacecraft_id": "1"
        }
        # Iterate on all the available product types:

        handler = LIL2NCFileHandler("filename", filename_info, extract_filetype_info(filetype_infos, "li_l2_af_nc"))

        dataset_id = make_dataid(name="test_dataset")
        with pytest.raises(KeyError):
            handler.get_dataset(dataset_id)

    def test_dataset_not_in_provided_dataset(self, filetype_infos):
        """Test loading of a dataset that is not provided."""
        filename_info = {
            "start_time": "0000",
            "end_time": "1000",
            "mission_prefix": "MT",
            "spacecraft_id": "1"
        }
        # Iterate on all the available product types:

        dataset_dict = {"name": "test_dataset"}

        handler = LIL2NCFileHandler("filename", filename_info, extract_filetype_info(filetype_infos, "li_l2_af_nc"))

        dataset_id = make_dataid(name="test_dataset")

        assert handler.get_dataset(dataset_id, ds_info=dataset_dict) is None

    def test_filename_infos(self, filetype_infos):
        """Test settings retrieved from filename."""
        filename_info = {
            "start_time": "20101112131415",
            "end_time": "20101112131416",
            "mission_prefix": "MT",
            "spacecraft_id": "1"

        }

        handler = LIL2NCFileHandler("filename", filename_info, extract_filetype_info(filetype_infos, "li_l2_af_nc"))

        # Start and end time should come from filename info:
        assert handler.start_time == "20101112131415"
        assert handler.end_time == "20101112131416"

        # internal vars should be initialized:
        assert handler.search_paths is not None
        assert handler.dataset_infos is not None

        # calling register_available_datasets again should not change things (early return)
        ds_infos_current = handler.dataset_infos.copy()
        handler.register_available_datasets()
        assert handler.dataset_infos == ds_infos_current

        # Should have some datasets:
        assert len(handler.provided_datasets) > 0

        # Sensor names should be just 'li'
        assert handler.sensor_names == {"li"}

        # check product type:
        assert handler.product_type == "2-AF"

    def test_platform_name_and_sensor(self, filetype_infos):
        """Test that plaform name and sensor attributes are set."""
        filename_info = {
            "start_time": "20101112131415",
            "end_time": "20101112131416",
            "mission_prefix": "MT",
            "spacecraft_id": "1"

        }

        handler = LIL2NCFileHandler("filename", filename_info, extract_filetype_info(filetype_infos, "li_l2_lfl_nc"))
        dsid = make_dataid(name="flash_duration")
        dset = handler.get_dataset(dsid)
        assert dset.attrs["platform_name"] == "Meteosat-12"
        assert dset.attrs["sensor"] == "li"

    def test_var_path_exists(self, filetype_infos):
        """Test variable_path_exists from li reader."""
        filename_info = {
            "start_time": "20101112131415",
            "end_time": "20101112131416",
            "mission_prefix": "MT",
            "spacecraft_id": "1"
        }

        handler = LIL2NCFileHandler("filename", filename_info, extract_filetype_info(filetype_infos, "li_l2_lef_nc"))

        # Check variable paths:
        assert handler.variable_path_exists("dummy") is False

        assert handler.variable_path_exists("state/processor/l1b_geolocation_warning") is False
        assert handler.variable_path_exists("data/l1b_geolocation_warning") is True
        assert handler.variable_path_exists("data/north/event_id") is True
        assert handler.variable_path_exists("data/none/event_id") is False
        assert handler.variable_path_exists("/attr") is False
        assert handler.variable_path_exists("data/l1b_geolocation_warning/dtype") is False
        assert handler.variable_path_exists("data/l1b_geolocation_warning/shape") is False
        assert handler.variable_path_exists("data/l1b_geolocation_warning/dimensions") is False

    def test_get_first_valid_variable(self, filetype_infos):
        """Test get_first_valid_variable from li reader."""
        filename_info = {
            "start_time": "20101112131415",
            "end_time": "20101112131416",
            "mission_prefix": "MT",
            "spacecraft_id": "1"
        }

        handler = LIL2NCFileHandler("filename", filename_info, extract_filetype_info(filetype_infos, "li_l2_lef_nc"))

        # Check variable paths:
        var1 = handler.get_first_valid_variable(["dummy/path", "data/north/detector_column"])
        var2 = handler.get_first_valid_variable(["dummy/path", "data/east/detector_column"])
        var3 = handler.get_first_valid_variable(["dummy/path", "data/south/detector_row"])
        var4 = handler.get_first_valid_variable(["dummy/path", "data/west/detector_row"])

        assert isinstance(var1, xr.DataArray)
        assert isinstance(var2, xr.DataArray)
        assert isinstance(var3, xr.DataArray)
        assert isinstance(var4, xr.DataArray)

        assert id(var1) != id(var2)
        assert id(var2) != id(var3)
        assert id(var3) != id(var4)

        mix1 = handler.get_first_valid_variable(["dummy/path",
                                                 "data/north/detector_column",
                                                 "data/east/detector_column",
                                                 "data/south/detector_row"])

        mix2 = handler.get_first_valid_variable(["dummy/path",
                                                 "data/west/detector_row",
                                                 "data/north/detector_column",
                                                 "data/east/detector_column",
                                                 "data/south/detector_row"])

        # first mix should give us var1 and the second one var4:
        assert id(mix1) == id(var1)
        assert id(mix2) == id(var4)

        # get the measured variables now:
        # Note that we must specify fill_value==None below otherwise
        # a new array is generated filling the invalid values:
        meas1 = handler.get_measured_variable("east/detector_column", fill_value=None)
        meas2 = handler.get_measured_variable("south/detector_row", fill_value=None)

        assert id(meas1) == id(var2)
        assert id(meas2) == id(var3)

        # We should have a fill value on those variables:
        assert var1.attrs.get("_FillValue") == 65535
        assert var2.attrs.get("_FillValue") == 65535

    def test_get_first_valid_variable_not_found(self, filetype_infos):
        """Test get_first_valid_variable from li reader if the variable is not found."""
        filename_info = {
            "start_time": "20101112131415",
            "end_time": "20101112131416",
            "mission_prefix": "MT",
            "spacecraft_id": "1"
        }

        handler = LIL2NCFileHandler("filename", filename_info, extract_filetype_info(filetype_infos, "li_l2_lef_nc"))

        with pytest.raises(KeyError):
            handler.get_first_valid_variable(["dummy/path", "data/test/test_var"])

    def test_available_datasets(self, filetype_infos):
        """Test available_datasets from li reader."""
        filename_info = {
            "start_time": "20101112131415",
            "end_time": "20101112131416",
            "mission_prefix": "MT",
            "spacecraft_id": "1"
        }
        handler = LIL2NCFileHandler("filename", filename_info, extract_filetype_info(filetype_infos, "li_l2_lef_nc"))

        # get current ds_infos. These should all be returned by the available_datasets
        ds_infos_to_compare = handler.dataset_infos.copy()

        # now add a dummy configured dataset to make sure that it is included in the available_datasets output
        ds_info_dummy = {"test": "test"}
        conf_ds_dummy = [(True, ds_info_dummy)]
        ds_infos_to_compare.insert(0, ds_info_dummy)

        assert ds_infos_to_compare == [ds[1] for ds in handler.available_datasets(configured_datasets=conf_ds_dummy)]

    def test_variable_scaling(self, filetype_infos):
        """Test automatic rescaling with offset and scale attributes."""
        filename_info = {
            "start_time": "20101112131415",
            "end_time": "20101112131416",
            "mission_prefix": "MT",
            "spacecraft_id": "1"
        }

        handler = LIL2NCFileHandler("filename", filename_info, extract_filetype_info(filetype_infos, "li_l2_lfl_nc"))

        # Get the raw variable without rescaling:
        vname = "latitude"
        rawlat = handler.get_measured_variable(vname)

        # Get the dataset without rescaling:
        dataset_info = {
            "name": vname,
            "variable_name": vname,
            "use_rescaling": False,
        }

        dataset_id = make_dataid(name=vname)
        lat_noscale = handler.get_dataset(dataset_id, dataset_info)
        assert np.all(lat_noscale.values == rawlat)

        # Now get the dataset with scaling:
        dataset_info["use_rescaling"] = True
        lat_scaled = handler.get_dataset(dataset_id, dataset_info)

        # By default we write data in the ranges [-88.3/0.0027, 88.3/0.0027] for latitude and longitude:
        assert abs(np.nanmax(lat_scaled.values) - 88.3) < 1e-2
        assert abs(np.nanmin(lat_scaled.values) + 88.3) < 1e-2

    def test_swath_coordinates(self, filetype_infos):
        """Test that swath coordinates are used correctly to assign coordinates to some datasets."""
        filename_info = {
            "start_time": "20101112131415",
            "end_time": "20101112131416",
            "mission_prefix": "MT",
            "spacecraft_id": "1"
        }
        handler = LIL2NCFileHandler("filename", filename_info, extract_filetype_info(filetype_infos, "li_l2_lfl_nc"))

        # Check latitude:
        dsid = make_dataid(name="latitude")
        dset = handler.get_dataset(dsid)
        assert "coordinates" not in dset.attrs

        # get_area_def should raise exception:
        with pytest.raises(NotImplementedError):
            handler.get_area_def(dsid)

        # Check radiance:
        dsid = make_dataid(name="radiance")
        dset = handler.get_dataset(dsid)
        assert "coordinates" in dset.attrs
        assert dset.attrs["coordinates"][0] == "longitude"
        assert dset.attrs["coordinates"][1] == "latitude"

        with pytest.raises(NotImplementedError):
            handler.get_area_def(dsid)

    def test_report_datetimes(self, filetype_infos):
        """Should report time variables as numpy datetime64 type and time durations as timedelta64."""
        filename_info = {
            "start_time": "20101112131415",
            "end_time": "20101112131416",
            "mission_prefix": "MT",
            "spacecraft_id": "1"
        }
        handler = LIL2NCFileHandler("filename", filename_info, extract_filetype_info(filetype_infos, "li_l2_le_nc"))

        # Check epoch_time:
        dsid = make_dataid(name="epoch_time_north_sector")
        dset = handler.get_dataset(dsid)
        assert dset.values.dtype == np.dtype("datetime64[ns]")

        # The default epoch_time should be 1.234 seconds after epoch:
        ref_time = np.datetime64(dt.datetime(2000, 1, 1, 0, 0, 1, 234000))
        assert np.all(dset.values == ref_time)

        # Check time_offset:
        dsid = make_dataid(name="time_offset_east_sector")
        dset = handler.get_dataset(dsid)
        assert dset.values.dtype == np.dtype("timedelta64[ns]")

        # The default time_offset should be: np.linspace(0.0, 1000.0, nobs)
        # but then we first multiply by 1e6 to generate us times:
        # Note that below no automatic transform to np.float64 is happening:
        nobs = dset.shape[0]
        ref_data = np.linspace(0.0, 1000.0, nobs).astype(np.float32)
        ref_data = (ref_data * 1e9).astype("timedelta64[ns]")

        # And not absolutely sure why, but we always get the timedelta in ns from the dataset:
        # ref_data = (ref_data).astype('timedelta64[ns]')

        assert np.all(dset.values == ref_data)

    def test_milliseconds_to_timedelta(self, filetype_infos):
        """Should covert milliseconds to timedelta."""
        filename_info = {
            "start_time": "20101112131415",
            "end_time": "20101112131416",
            "mission_prefix": "MT",
            "spacecraft_id": "1"
        }
        handler = LIL2NCFileHandler("filename", filename_info, extract_filetype_info(filetype_infos, "li_l2_lfl_nc"))

        # Check flash_duration:
        dsid = make_dataid(name="flash_duration")
        dset = handler.get_dataset(dsid)
        assert dset.values.dtype == np.dtype("timedelta64[ns]")

        nobs = dset.shape[0]
        ref_data = np.linspace(0, 1000, nobs).astype("u2")
        ref_data = (ref_data * 1e6).astype("timedelta64[ns]")

        assert np.all(dset.values == ref_data)

    def test_apply_accumulate_index_offset(self, filetype_infos):
        """Should accumulate index offsets."""
        filename_info = {
            "start_time": "20101112131415",
            "end_time": "20101112131416",
            "mission_prefix": "MT",
            "spacecraft_id": "1"
        }
        handler = LIL2NCFileHandler("filename", filename_info, extract_filetype_info(filetype_infos, "li_l2_le_nc"))

        # Check time offset:
        dsid = make_dataid(name="l1b_chunk_offsets_north_sector")
        dset = handler.get_dataset(dsid)

        nobs = dset.shape[0]
        ref_data = (np.arange(nobs)).astype("u4")
        # check first execution without offset
        assert np.all(dset.values == ref_data)
        # check that the offset is being stored
        assert handler.current_ds_info["__index_offset"] == 123

        # check execution with offset value
        # this simulates the case where we are loading this variable from multiple files and concatenating it
        dset = handler.get_dataset(dsid, handler.current_ds_info)
        assert np.all(dset.values == ref_data + 123)

    def test_combine_info(self, filetype_infos):
        """Test overridden combine_info."""
        filename_info = {
            "start_time": "20101112131415",
            "end_time": "20101112131416",
            "mission_prefix": "MT",
            "spacecraft_id": "1"
        }
        handler = LIL2NCFileHandler("filename", filename_info, extract_filetype_info(filetype_infos, "li_l2_le_nc"))

        # get a dataset including the index_offset in the ds_info
        dsid = make_dataid(name="l1b_chunk_offsets_north_sector")
        ds_info = {"name": "l1b_chunk_offsets_north_sector",
                   "variable_name": "l1b_chunk_offsets",
                   "sector_name": "north",
                   "__index_offset": 1000,
                   "accumulate_index_offset": "{sector_name}/l1b_window"}
        dset = handler.get_dataset(dsid, ds_info=ds_info)
        handler.combine_info([dset.attrs])
        # combine_info should have removed the index_offset key from the ds_info passed to get_dataset
        assert "__index_offset" not in ds_info
        # and reset the current_ds_info dict, in order to avoid failures if we call combine_info again
        assert handler.current_ds_info is None

    def test_coordinates_projection(self, filetype_infos):
        """Should automatically generate lat/lon coords from projection data."""
        filename_info = {
            "start_time": "20101112131415",
            "end_time": "20101112131416",
            "mission_prefix": "MT",
            "spacecraft_id": "1"
        }
        handler = LIL2NCFileHandler("filename", filename_info, extract_filetype_info(filetype_infos, "li_l2_af_nc"),
                                    with_area_definition=False)

        dsid = make_dataid(name="flash_accumulation")
        dset = handler.get_dataset(dsid)
        assert "coordinates" in dset.attrs

        assert dset.attrs["coordinates"][0] == "longitude"
        assert dset.attrs["coordinates"][1] == "latitude"

        with pytest.raises(NotImplementedError):
            handler.get_area_def(dsid)

        handler = LIL2NCFileHandler("filename", filename_info, extract_filetype_info(filetype_infos, "li_l2_afr_nc"),
                                    with_area_definition=False)

        dsid = make_dataid(name="flash_radiance")
        dset = handler.get_dataset(dsid)
        assert "coordinates" in dset.attrs

        assert dset.attrs["coordinates"][0] == "longitude"
        assert dset.attrs["coordinates"][1] == "latitude"

        handler = LIL2NCFileHandler("filename", filename_info, extract_filetype_info(filetype_infos, "li_l2_afa_nc"),
                                    with_area_definition=False)

        dsid = make_dataid(name="accumulated_flash_area")
        dset = handler.get_dataset(dsid)
        assert "coordinates" in dset.attrs

        assert dset.attrs["coordinates"][0] == "longitude"
        assert dset.attrs["coordinates"][1] == "latitude"

    def test_generate_coords_on_accumulated_prods(self, filetype_infos):
        """Test daskified generation of coords."""
        accumulated_products = ["li_l2_af_nc", "li_l2_afr_nc", "li_l2_afa_nc"]
        coordinate_datasets = ["longitude", "latitude"]
        filename_info = {
            "start_time": "20101112131415",
            "end_time": "20101112131416",
            "mission_prefix": "MT",
            "spacecraft_id": "1"
        }

        for accum_prod in accumulated_products:
            for ds_name in coordinate_datasets:
                handler = LIL2NCFileHandler("filename", filename_info,
                                            extract_filetype_info(filetype_infos, accum_prod))
                dsid = make_dataid(name=ds_name)
                dset = handler.get_dataset(dsid)
                # Check dataset type
                assert isinstance(dset, xr.DataArray)
                vals = dset.values
                assert vals is not None

    def test_generate_coords_on_lon_lat(self, filetype_infos):
        """Test getting lon/lat dataset on accumulated product."""
        accumulated_products = ["li_l2_af_nc", "li_l2_afr_nc", "li_l2_afa_nc"]
        coordinate_datasets = ["longitude", "latitude"]
        filename_info = {
            "start_time": "20101112131415",
            "end_time": "20101112131416",
            "mission_prefix": "MT",
            "spacecraft_id": "1"
        }

        for accum_prod in accumulated_products:
            for ds_name in coordinate_datasets:
                handler = LIL2NCFileHandler("filename", filename_info,
                                            extract_filetype_info(filetype_infos, accum_prod))
                dsid = make_dataid(name=ds_name)
                handler.generate_coords_from_scan_angles = mock.MagicMock(
                    side_effect=handler.generate_coords_from_scan_angles)
                handler.get_dataset(dsid)
                assert handler.generate_coords_from_scan_angles.called

    def test_generate_coords_inverse_proj(self, filetype_infos):
        """Test inverse_projection execution delayed until .values is called on the dataset."""
        accumulated_products = ["li_l2_af_nc", "li_l2_afr_nc", "li_l2_afa_nc"]
        coordinate_datasets = ["longitude", "latitude"]
        filename_info = {
            "start_time": "20101112131415",
            "end_time": "20101112131416",
            "mission_prefix": "MT",
            "spacecraft_id": "1"
        }

        for accum_prod in accumulated_products:
            for ds_name in coordinate_datasets:
                handler = LIL2NCFileHandler("filename", filename_info,
                                            extract_filetype_info(filetype_infos, accum_prod))
                dsid = make_dataid(name=ds_name)
                handler.inverse_projection = mock.MagicMock(side_effect=handler.inverse_projection)
                dset = handler.get_dataset(dsid)
                assert not handler.inverse_projection.called
                vals = dset.values
                assert vals is not None
                assert handler.inverse_projection.called

    def test_generate_coords_not_called_on_non_coord_dataset(self, filetype_infos):
        """Test that the method is not called when getting non-coord dataset."""
        handler = self.generate_coords(filetype_infos, "li_l2_af_nc", "flash_accumulation")
        assert not handler.generate_coords_from_scan_angles.called

    def test_generate_coords_not_called_on_non_accum_dataset(self, filetype_infos):
        """Test that the method is not called when getting non-accum dataset."""
        handler = self.generate_coords(filetype_infos, "li_l2_lef_nc", "latitude_north_sector")
        assert not handler.generate_coords_from_scan_angles.called

    def generate_coords(self, filetype_infos, file_type_name, variable_name):
        """Generate file handler and mimic coordinate generator call."""
        filename_info = {
            "start_time": "20101112131415",
            "end_time": "20101112131416",
            "mission_prefix": "MT",
            "spacecraft_id": "1"
        }
        handler = LIL2NCFileHandler("filename", filename_info, extract_filetype_info(filetype_infos, file_type_name))
        dsid = make_dataid(name=variable_name)
        handler.generate_coords_from_scan_angles = mock.MagicMock(
            side_effect=handler.generate_coords_from_scan_angles)
        handler.get_dataset(dsid)
        return handler

    def test_generate_coords_called_once(Self, filetype_infos):
        """Test that the method is called only once."""
        filename_info = {
            "start_time": "20101112131415",
            "end_time": "20101112131416",
            "mission_prefix": "MT",
            "spacecraft_id": "1"
        }
        handler = LIL2NCFileHandler("filename", filename_info, extract_filetype_info(filetype_infos, "li_l2_af_nc"))
        # check internal variable is empty
        assert len(handler.internal_variables) == 0
        coordinate_datasets = ["longitude", "latitude"]
        handler.generate_coords_from_scan_angles = mock.MagicMock(side_effect=handler.generate_coords_from_scan_angles)

        for ds_name in coordinate_datasets:
            dsid = make_dataid(name=ds_name)
            dset = handler.get_dataset(dsid)
            # Check dataset type
            assert isinstance(dset, xr.DataArray)
            assert len(handler.internal_variables) == 2
            assert handler.generate_coords_from_scan_angles.called

    def test_coords_generation(self, filetype_infos):
        """Compare daskified coords generation results with non-daskified."""
        products = ["li_l2_af_nc",
                    "li_l2_afr_nc",
                    "li_l2_afa_nc"]

        filename_info = {
            "start_time": "20101112131415",
            "end_time": "20101112131416",
            "mission_prefix": "MT",
            "spacecraft_id": "1"
        }
        for prod in products:
            handler = LIL2NCFileHandler("filename", filename_info, extract_filetype_info(filetype_infos, prod))

            # Get azimuth/elevation arrays from handler
            azimuth = handler.get_measured_variable(handler.swath_coordinates["azimuth"])
            azimuth = handler.apply_use_rescaling(azimuth)

            elevation = handler.get_measured_variable(handler.swath_coordinates["elevation"])
            elevation = handler.apply_use_rescaling(elevation)
            # Initialize proj_dict
            proj_var = handler.swath_coordinates["projection"]
            geos_proj = handler.get_measured_variable(proj_var, fill_value=None)
            major_axis = float(geos_proj.attrs["semi_major_axis"])
            point_height = 35786400.0  # float(geos_proj.attrs["perspective_point_height"])
            inv_flattening = float(geos_proj.attrs["inverse_flattening"])
            lon_0 = float(geos_proj.attrs["longitude_of_projection_origin"])
            sweep = str(geos_proj.attrs["sweep_angle_axis"])
            proj_dict = {"a": major_axis,
                         "lon_0": lon_0,
                         "h": point_height,
                         "rf": inv_flattening,
                         "proj": "geos",
                         "units": "m",
                         "sweep": sweep}

            # Compute reference values
            projection = Proj(proj_dict)
            azimuth_vals = azimuth.values * point_height
            elevation_vals = elevation.values * point_height
            azimuth_vals *= -1
            lon_ref, lat_ref = projection(azimuth_vals, elevation_vals, inverse=True)

            # Convert lon_ref, lat_ref to a np.float32
            lon_ref = lon_ref.astype(np.float32)
            lat_ref = lat_ref.astype(np.float32)

            handler.generate_coords_from_scan_angles()
            lon = handler.internal_variables["longitude"].values
            lat = handler.internal_variables["latitude"].values

            # Compare the arrays, should be the same:
            np.testing.assert_equal(lon, lon_ref)
            np.testing.assert_equal(lat, lat_ref)

    def test_coords_and_grid_consistency(self, filetype_infos):
        """Compare computed latlon coords for 1-d version with latlon from areadef as for the gridded version."""
        filename_info = {
            "start_time": "20101112131415",
            "end_time": "20101112131416",
            "mission_prefix": "MT",
            "spacecraft_id": "1"
        }
        handler = LIL2NCFileHandler("filename", filename_info, extract_filetype_info(filetype_infos, "li_l2_af_nc"),
                                    with_area_definition=True)

        # Get cols/rows arrays from handler
        x = handler.get_measured_variable(handler.swath_coordinates["azimuth"])
        y = handler.get_measured_variable(handler.swath_coordinates["elevation"])
        cols = x.astype(int) - 1
        rows = (LI_GRID_SHAPE[0] - y.astype(int))

        # compute lonlat from 1-d coords generation (called when with_area_definition==False)
        handler.generate_coords_from_scan_angles()
        lon = handler.internal_variables["longitude"].values
        lat = handler.internal_variables["latitude"].values

        # compute lonlat from 2-d areadef
        dsid = make_dataid(name="flash_accumulation")
        area_def = handler.get_area_def(dsid)
        lon_areadef, lat_areadef = area_def.get_lonlat_from_array_coordinates(cols, rows)

        np.testing.assert_allclose(lon, lon_areadef, rtol=1e-3)
        np.testing.assert_allclose(lat, lat_areadef, rtol=1e-3)

    def test_get_area_def_acc_products(self, filetype_infos):
        """Test retrieval of area def for accumulated products."""
        filename_info = {
            "start_time": "20101112131415",
            "end_time": "20101112131416",
            "mission_prefix": "MT",
            "spacecraft_id": "1"
        }
        handler = LIL2NCFileHandler("filename", filename_info, extract_filetype_info(filetype_infos, "li_l2_af_nc"),
                                    with_area_definition=True)

        dsid = make_dataid(name="flash_accumulation")

        area_def = handler.get_area_def(dsid)
        assert area_def.shape == LI_GRID_SHAPE

        # Should throw for non-gridded variables:
        with pytest.raises(NotImplementedError):
            handler.get_area_def(make_dataid(name="accumulation_offsets"))

    def test_get_area_def_non_acc_products(self, filetype_infos):
        """Test retrieval of area def for non-accumulated products."""
        filename_info = {
            "start_time": "20101112131415",
            "end_time": "20101112131416",
            "mission_prefix": "MT",
            "spacecraft_id": "1"
        }
        handler = LIL2NCFileHandler("filename", filename_info, extract_filetype_info(filetype_infos, "li_l2_lgr_nc"),
                                    with_area_definition=True)
        # Should throw for non-accum products:
        with pytest.raises(NotImplementedError):
            handler.get_area_def(make_dataid(name="radiance"))

    @staticmethod
    def param_provider(_filename, filename_info, _fileype_info):
        """Provide parameters."""

        def write_flash_accum(_vname, _ocname, _settings):
            """Write the flash accumulation array."""
            return np.arange(1234, dtype=np.float32) + 0.5

        # We return the settings we want to use here to generate our custom/fixed product content:
        return {
            "num_obs": 1234,
            "providers": {
                "flash_accumulation": write_flash_accum,
            }
        }

    def test_without_area_def(self, filetype_infos):
        """Test accumulated products data array without area definition."""
        # without area definition
        filename_info = {
            "start_time": "20101112131415",
            "end_time": "20101112131416",
            "mission_prefix": "MT",
            "spacecraft_id": "1"
        }
        handler_without_area_def = LIL2NCFileHandler(
            "filename", filename_info, extract_filetype_info(filetype_infos, "li_l2_af_nc"), with_area_definition=False)

        dsid = make_dataid(name="flash_accumulation")

        # Keep the data array:
        data = handler_without_area_def.get_dataset(dsid).values
        assert data.shape == (1234,)

    def test_with_area_def(self, filetype_infos):
        """Test accumulated products data array with area definition."""
        handler = self.handler_with_area(filetype_infos, "li_l2_af_nc")
        dsid = make_dataid(name="flash_accumulation")
        # Retrieve the 2D array:
        arr = handler.get_dataset(dsid).values
        assert arr.shape == LI_GRID_SHAPE

    def test_get_on_fci_grid_exc(self, filetype_infos):
        """Test the execution of the get_on_fci_grid function for an accumulated gridded variable."""
        handler = self.handler_with_area(filetype_infos, "li_l2_af_nc")
        handler.get_array_on_fci_grid = mock.MagicMock(side_effect=handler.get_array_on_fci_grid)
        dsid = make_dataid(name="flash_accumulation")
        handler.get_dataset(dsid)
        assert handler.get_array_on_fci_grid.called

    def test_get_on_fci_grid_exc_non_grid(self, filetype_infos):
        """Test the non-execution of the get_on_fci_grid function for an accumulated non-gridded variable."""
        handler = self.handler_with_area(filetype_infos, "li_l2_af_nc")
        handler.get_array_on_fci_grid = mock.MagicMock(side_effect=handler.get_array_on_fci_grid)
        dsid = make_dataid(name="accumulation_offsets")
        handler.get_dataset(dsid)
        assert not handler.get_array_on_fci_grid.called

    def test_get_on_fci_grid_exc_non_accum(self, filetype_infos):
        """Test the non-execution of the get_on_fci_grid function for a non-accumulated variable."""
        handler = self.handler_with_area(filetype_infos, "li_l2_lef_nc")
        handler.get_array_on_fci_grid = mock.MagicMock(side_effect=handler.get_array_on_fci_grid)
        dsid = make_dataid(name="radiance_north_sector")
        handler.get_dataset(dsid)
        assert not handler.get_array_on_fci_grid.called

    def test_with_area_def_vars_with_no_pattern(self, filetype_infos):
        """Test accumulated products variable with no patterns and with area definition."""
        handler = self.handler_with_area(filetype_infos, "li_l2_af_nc")
        # variable with no patterns
        dsid = make_dataid(name="accumulation_offsets")
        assert handler.get_dataset(dsid).shape == (1,)

    def handler_with_area(self, filetype_infos, product_name):
        """Create handler with area definition."""
        filename_info = {
            "start_time": "20101112131415",
            "end_time": "20101112131416",
            "mission_prefix": "MT",
            "spacecraft_id": "1"
        }
        # Note: we need a test param provider here to ensure we write the same values for both handlers below:
        FakeLIFileHandlerBase.schema_parameters = TestLIL2.param_provider
        # with area definition
        handler = LIL2NCFileHandler("filename", filename_info, extract_filetype_info(filetype_infos, product_name),
                                    with_area_definition=True)
        return handler

    def test_with_area_def_pixel_placement(self, filetype_infos):
        """Test the placements of pixel value with area definition."""
        filename_info = {
            "start_time": "20101112131415",
            "end_time": "20101112131416",
            "mission_prefix": "MT",
            "spacecraft_id": "1"
        }
        # with area definition
        FakeLIFileHandlerBase.schema_parameters = TestLIL2.param_provider

        handler = LIL2NCFileHandler("filename", filename_info, extract_filetype_info(filetype_infos, "li_l2_af_nc"),
                                    with_area_definition=True)
        dsid = make_dataid(name="flash_accumulation")

        # Retrieve the 2D array:a
        arr = handler.get_dataset(dsid).values

        # Retrieve the x/y coordinates:
        xarr = handler.get_measured_variable("x").values.astype(int)
        yarr = handler.get_measured_variable("y").values.astype(int)

        handler_without_area_def = LIL2NCFileHandler(
            "filename", filename_info, extract_filetype_info(filetype_infos, "li_l2_af_nc"), with_area_definition=False)

        FakeLIFileHandlerBase.schema_parameters = None

        # prepare reference array
        data = handler_without_area_def.get_dataset(dsid).values
        ref_arr = np.empty(LI_GRID_SHAPE, dtype=arr.dtype)
        ref_arr[:] = 0
        rows = (LI_GRID_SHAPE[0] - yarr)
        cols = xarr - 1
        for n_entry in range(len(data)):
            ref_arr[rows[n_entry], cols[n_entry]] += data[n_entry]
        ref_arr = np.where(ref_arr > 0, ref_arr, np.nan)

        # Check all nan values are at the same locations:
        assert np.all(np.isnan(arr) == np.isnan(ref_arr))

        # Check all finite values are the same:
        assert np.all(arr[np.isfinite(arr)] == ref_arr[np.isfinite(ref_arr)])



# Copyright (c) 2017-2025 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unit testing the enhancement decoration functions."""

import dask.array as da
import numpy as np
import pytest
import xarray as xr

from satpy.enhancements.wrappers import on_dask_array, on_separate_bands, using_map_blocks

from .utils import create_ch1, create_ch2, create_rgb, run_and_check_enhancement


def identical_decorator(func):
    """Decorate but do nothing."""
    return func


class TestEnhancementsConvolution:
    """Class for testing decorators in satpy.enhancements.wrappers module."""

    def setup_method(self):
        """Create test data used by every test."""
        self.ch1 = create_ch1()
        self.ch2 = create_ch2()
        self.rgb = create_rgb()

    @pytest.mark.parametrize(
        ("decorator", "exp_call_cls"),
        [
            (identical_decorator, xr.DataArray),
            (on_dask_array, da.Array),
            (using_map_blocks, np.ndarray),
        ],
    )
    @pytest.mark.parametrize("input_data_name", ["ch1", "ch2", "rgb"])
    def test_enhancement_decorators(self, input_data_name, decorator, exp_call_cls):
        """Test the utility decorators."""

        def _enh_func(img):
            def _calc_func(data):
                assert isinstance(data, exp_call_cls)
                return data

            decorated_func = decorator(_calc_func)
            return decorated_func(img.data)

        in_data = getattr(self, input_data_name)
        exp_data = in_data.values
        if "bands" not in in_data.coords:
            exp_data = exp_data[np.newaxis, :, :]
        run_and_check_enhancement(_enh_func, in_data, exp_data)

    def tearDown(self):
        """Clean up."""


def test_on_separate_bands():
    """Test the `on_separate_bands` decorator."""

    def func(array, index, gain=2):
        return xr.DataArray(np.ones(array.shape, dtype=array.dtype) * index * gain,
                            coords=array.coords, dims=array.dims, attrs=array.attrs)

    separate_func = on_separate_bands(func)
    arr = xr.DataArray(np.zeros((3, 10, 10)), dims=["bands", "y", "x"], coords={"bands": ["R", "G", "B"]})
    assert separate_func(arr).shape == arr.shape
    assert all(separate_func(arr, gain=1).values[:, 0, 0] == [0, 1, 2])


def test_using_map_blocks():
    """Test the `using_map_blocks` decorator."""

    def func(np_array, block_info=None):
        value = block_info[0]["chunk-location"][-1]
        return np.ones(np_array.shape) * value

    map_blocked_func = using_map_blocks(func)
    arr = xr.DataArray(da.zeros((3, 10, 10), dtype=int, chunks=5), dims=["bands", "y", "x"])
    res = map_blocked_func(arr)
    assert res.shape == arr.shape
    assert res[0, 0, 0].compute() != res[0, 9, 9].compute()


def test_on_dask_array():
    """Test the `on_dask_array` decorator."""

    def func(dask_array):
        if not isinstance(dask_array, da.core.Array):
            pytest.fail("Array is not a dask array")
        return dask_array

    dask_func = on_dask_array(func)
    arr = xr.DataArray(da.zeros((3, 10, 10), dtype=int, chunks=5), dims=["bands", "y", "x"])
    res = dask_func(arr)
    assert res.shape == arr.shape



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2025- Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""NWC SAF GEO v2021 HRW.

This module contains the :class:`NWCSAFGEOHRWFileHandler` file handler for the High Resolution Winds (HRW)
data produced with NWC SAF GEO.

There are 37 different quantities in the files, and predicted trajectories for each
observation. Reading of the trajectory data are not currently supported.

The data in the files are grouped separately for each original imaging channel, which there are up to
seven. The number of channels depends on the configuration of NWC SAF GEO software.

By default the reader treats each imaging channel separately, and thus lists
7 (channels) * 37 (variables) = 259 distinct datasets::

    import pprint
    from satpy import Scene

    filename = "S_NWC_HRW_MSG3_MSG-N-BS_20250206T130000Z.nc"
    scn = Scene(reader="nwcsaf-geo", filenames=[filename])
    pprint.pprint(scn.available_dataset_names())

This prints all the available datasets. The truncated output of this is::

    ['wind_hrvis_air_pressure',
     'wind_hrvis_air_pressure_correction',
     'wind_hrvis_air_pressure_error',
     'wind_hrvis_air_pressure_nwp_at_best_fit_level',
     'wind_hrvis_air_temperature',
     ...
     'wind_wv073_wind_speed',
     'wind_wv073_wind_speed_difference_nwp_at_amv_level',
     'wind_wv073_wind_speed_difference_nwp_at_best_fit_level',
     'wind_wv073_wind_speed_nwp_at_amv_level',
     'wind_wv073_wind_speed_nwp_at_best_fit_level']

The channel name is used as a prefix for the datasets.

It is also possible to merge all of these channel-by-channel datasets with a reader key-word argument::

    scn = Scene(reader="nwcsaf-geo", filenames=[filename], reader_kwargs={"merge_channels": True})
    pprint.pprint(scn.available_dataset_names())

Full list of the printed datasets is::

    ['air_pressure',
     'air_pressure_correction',
     'air_pressure_error',
     'air_pressure_nwp_at_best_fit_level',
     'air_temperature',
     'cloud_type',
     'correlation',
     'correlation_test',
     'height_assignment_method',
     'latitude',
     'latitude_increment',
     'longitude',
     'longitude_increment',
     'number_of_winds',
     'orographic_index',
     'previous_wind_idx',
     'quality_index_iwwg_value',
     'quality_index_with_forecast',
     'quality_index_without_forecast',
     'quality_test',
     'segment_x',
     'segment_x_pix',
     'segment_y',
     'segment_y_pix',
     'tracer_correlation_method',
     'tracer_type',
     'wind_from_direction',
     'wind_from_direction_difference_nwp_at_amv_level',
     'wind_from_direction_difference_nwp_at_best_fit_level',
     'wind_from_direction_nwp_at_amv_level',
     'wind_from_direction_nwp_at_best_fit_level',
     'wind_idx',
     'wind_speed',
     'wind_speed_difference_nwp_at_amv_level',
     'wind_speed_difference_nwp_at_best_fit_level',
     'wind_speed_nwp_at_amv_level',
     'wind_speed_nwp_at_best_fit_level']

"""

import datetime as dt
import logging
from abc import ABCMeta, abstractmethod
from contextlib import suppress

import dask.array as da
import h5py
import numpy as np
import xarray as xr
from packaging.version import Version

from satpy.readers.core.file_handlers import BaseFileHandler
from satpy.readers.nwcsaf_nc import PLATFORM_NAMES, SENSOR, read_nwcsaf_time
from satpy.utils import get_chunk_size_limit

logger = logging.getLogger(__name__)

CHUNK_SIZE = get_chunk_size_limit()

FIRST_V2025_ALGORITHM_VERSION = Version("7.0")

PRE_V2025_SEVIRI_WIND_CHANNELS = [
    "wind_hrvis",
    "wind_ir108",
    "wind_ir120",
    "wind_vis06",
    "wind_vis08",
    "wind_wv062",
    "wind_wv073",
]

POST_V2025_DATASETS = [
    "air_pressure",
    "air_pressure_correction",
    "air_pressure_error",
    "air_pressure_nwp_at_best_fit_level",
    "air_temperature",
    "barometric_altitude_in_hectofeet",
    "cloud_type",
    "correlation",
    "correlation_test",
    "height_assignment_method",
    "latitude",
    "latitude_increment",
    "longitude",
    "longitude_increment",
    "number_of_trajectory_points",
    "number_of_winds",
    "orographic_index",
    "quality_index_iwwg_value",
    "quality_index_with_forecast",
    "quality_index_without_forecast",
    "quality_test",
    "satellite_channel",
    "segment_x",
    "segment_x_pix",
    "segment_y",
    "segment_y_pix",
    "tracer_correlation_method",
    "tracer_type",
    "wind_from_direction",
    "wind_from_direction_difference_nwp_at_amv_level",
    "wind_from_direction_difference_nwp_at_best_fit_level",
    "wind_from_direction_nwp_at_amv_level",
    "wind_from_direction_nwp_at_best_fit_level",
    "wind_id",
    "wind_prev_id",
    "wind_speed",
    "wind_speed_difference_nwp_at_amv_level",
    "wind_speed_difference_nwp_at_best_fit_level",
    "wind_speed_nwp_at_amv_level",
    "wind_speed_nwp_at_best_fit_level",
]

# Source: NWC/CDOP3/GEO/AEMET/SW/DOF
DATASET_UNITS = {
    "air_pressure": "Pa",
    "air_pressure_correction": "Pa",
    "air_pressure_error": "Pa",
    "air_pressure_nwp_at_best_fit_level": "Pa",
    "air_temperature": "K",
    "cloud_type": "1",
    "correlation": "%",
    "correlation_test": "1",
    "height_assignment_method": "1",
    "latitude": "degree_north",
    "latitude_increment": "degree_north",
    "longitude": "degree_east",
    "longitude_increment": "degree_east",
    "number_of_winds": "1",
    "orographic_index": "1",
    "previous_wind_idx": "1",
    "quality_index_iwwg_value": "%",
    "quality_index_with_forecast": "%",
    "quality_index_without_forecast": "%",
    "quality_test": "1",
    "segment_x": "m",
    "segment_x_pix": "1",
    "segment_y": "m",
    "segment_y_pix": "1",
    "tracer_correlation_method": "1",
    "tracer_type": "1",
    "wind_from_direction": "degree",
    "wind_from_direction_difference_nwp_at_amv_level": "degree",
    "wind_from_direction_difference_nwp_at_best_fit_level": "degree",
    "wind_from_direction_nwp_at_amv_level": "degree",
    "wind_from_direction_nwp_at_best_fit_level": "degree",
    "wind_idx": "1",
    "wind_speed": "m/s",
    "wind_speed_difference_nwp_at_amv_level": "m/s",
    "wind_speed_difference_nwp_at_best_fit_level": "m/s",
    "wind_speed_nwp_at_amv_level": "m/s",
    "wind_speed_nwp_at_best_fit_level": "m/s"
}


def NWCSAFGEOHRWFileHandler(filename, filename_info, filetype_info, merge_channels=False):
    """Return NWC SAF HRW file handler for a specific version.

    The name mimics the old file handler class name.
    """
    with h5py.File(filename, "r") as fid:
        version = Version(fid.attrs["product_algorithm_version"].astype(str))

    if version < FIRST_V2025_ALGORITHM_VERSION:
        return NWCSAFHRWPreV2025FileHandler(filename, filename_info, filetype_info, merge_channels=merge_channels)
    else:
        return NWCSAFHRWV2025FileHandler(filename, filename_info, filetype_info)


class NWCSAFHRWBase(BaseFileHandler, metaclass=ABCMeta):
    """A file handler class for NWC SAF GEO HRW files."""

    def __init__(self, filename, filename_info, filetype_info, merge_channels=False):
        """Initialize the file handler."""
        super().__init__(filename, filename_info, filetype_info)

        self.h5f = h5py.File(self.filename, "r")
        self.filename_info = filename_info
        self.filetype_info = filetype_info
        self.merge_channels = merge_channels
        self.platform_name = PLATFORM_NAMES.get(self.h5f.attrs["satellite_identifier"].astype(str))
        self.sensor = SENSOR.get(self.platform_name, "seviri")
        self.lons = {}
        self.lats = {}
        # Imaging period, which is set after reading any data, and used to calculate end time
        self.period = None

        # The resolution is given in kilometers, convert to meters
        self.resolution = 1000 * self.h5f.attrs["spatial_resolution"].item()

    def __del__(self):
        """Close file handlers when we are done."""
        with suppress(OSError):
            self.h5f.close()

    @abstractmethod
    def available_datasets(self, configured_datasets=None):
        """Form the names for the available datasets."""

    def _measurand_ds_info(self, prefix, measurand):
        ds_info = {
            "file_type": self.filetype_info["file_type"],
            "resolution": self.resolution,
            "name": prefix + measurand,
        }
        if measurand not in ("longitude", "latitude"):
            ds_info["coordinates"] = (prefix + "longitude", prefix + "latitude")
        if measurand == "longitude":
            ds_info["standard_name"] = "longitude"
        if measurand == "latitude":
            ds_info["standard_name"] = "latitude"

        return ds_info

    @abstractmethod
    def get_dataset(self, key, info):
        """Load a dataset."""

    def _read_merged_dataset(self, dataset_key):
        """Read a dataset merged from every channel."""
        dataset_name = dataset_key["name"]

        data = []
        collect_coords = True
        if "merged" in self.lons:
            collect_coords = False
        for channel in PRE_V2025_SEVIRI_WIND_CHANNELS:
            if collect_coords:
                self._read_channel_coordinates(channel)
                self._append_merged_coordinates(channel)
            if self.period is None:
                self.period = self.h5f[channel].attrs["time_period"].item()
            try:
                data.append(self.h5f[channel][dataset_name])
            except ValueError:
                logger.warning("Reading %s is not supported.", dataset_name)

        units = DATASET_UNITS[dataset_name]

        return self._create_xarray(
            data,
            dataset_name,
            {"units": units},
            "merged"
        )

    def _append_merged_coordinates(self, channel):
        if "merged" not in self.lons:
            self.lons["merged"] = []
            self.lats["merged"] = []
        self.lons["merged"].append(self.lons[channel])
        self.lats["merged"].append(self.lats[channel])

    def _create_xarray(self, data, dataset_name, attrs, channel):
        lons = self.lons[channel]
        lats = self.lats[channel]
        prefix = channel + "_"
        if channel == "merged":
            data = np.concat(data)
            lons = np.concat(lons)
            lats = np.concat(lats)
            prefix = ""

        xr_data = xr.DataArray(
            da.from_array(data, chunks=CHUNK_SIZE),
            name=dataset_name,
            dims=["y"],
            attrs=attrs,
        )
        xr_data[prefix + "longitude"] = ("y", lons)
        xr_data[prefix + "latitude"] = ("y", lats)

        return xr_data

    @abstractmethod
    def _read_dataset(self, dataset_key):
        """Read a dataset."""

    @property
    def start_time(self):
        """Get the start time."""
        return read_nwcsaf_time(self.h5f.attrs["nominal_product_time"])

    @property
    def end_time(self):
        """Get the end time."""
        if self.period is None:
            return self.start_time
        return self.start_time + dt.timedelta(minutes=self.period)


class NWCSAFHRWPreV2025FileHandler(NWCSAFHRWBase):
    """File handler for NWC SAF HRW files generated with versions oldert than v2025."""

    def get_dataset(self, key, info):
        """Load a dataset."""
        logger.debug("Reading %s.", key["name"])
        if self.merge_channels:
            data = self._read_merged_dataset(key)
        else:
            data = self._read_dataset(key)
        data.attrs.update(info)

        return data

    def _read_dataset(self, dataset_key):
        """Read a dataset."""
        dataset_name = dataset_key["name"]
        channel, measurand = self._get_channel_and_measurand(dataset_name)
        self._read_channel_coordinates(channel)
        self._get_period(channel)

        try:
            data = self.h5f[channel][measurand]
            attrs = {"units": DATASET_UNITS[measurand]}
        except ValueError:
            logger.warning("Reading %s is not supported.", dataset_name)

        return self._create_xarray(data, dataset_name, attrs, channel)

    @staticmethod
    def _get_channel_and_measurand(dataset_name):
        key_parts = dataset_name.split("_")
        channel = "_".join(key_parts[:2])
        measurand = "_".join(key_parts[2:])

        return channel, measurand

    def _read_channel_coordinates(self, channel):
        if channel not in self.lons:
            self.lons[channel] = self.h5f[channel]["longitude"]
            self.lats[channel] = self.h5f[channel]["latitude"]

    def available_datasets(self, configured_datasets=None):
        """Form the names for the available datasets."""
        for channel in PRE_V2025_SEVIRI_WIND_CHANNELS:
            prefix = self._get_channel_prefix(channel)
            dset = self.h5f[channel]
            for measurand in dset.dtype.fields.keys():
                if measurand == "trajectory":
                    continue
                ds_info = self._measurand_ds_info(prefix, measurand)
                yield True, ds_info
            if self.merge_channels:
                break

    def _get_channel_prefix(self, channel):
        if self.merge_channels:
            return ""
        return channel + "_"

    def _get_period(self, channel):
        if self.period is None:
            self.period = self.h5f[channel].attrs["time_period"].item()


class NWCSAFHRWV2025FileHandler(NWCSAFHRWBase):
    """File handler for NWC SAF HRW files generated with versions oldert than v2025."""

    def get_dataset(self, key, info):
        """Load a dataset."""
        logger.debug("Reading %s.", key["name"])
        data = self._read_dataset(key)
        data.attrs.update(info)

        return data

    def _read_dataset(self, dataset_key):
        """Read a dataset."""
        dataset_name = dataset_key["name"]
        channel, measurand = self._get_channel_and_measurand(dataset_name)
        self._read_channel_coordinates(channel)
        self._get_period(channel)

        try:
            data = self.h5f[measurand]
            if data.dtype == np.double:
                data = np.where(data == data.attrs["_FillValue"], np.nan, data)
            attrs = dict(self.h5f[measurand].attrs)
        except ValueError:
            logger.warning("Reading %s is not supported.", dataset_name)

        return self._create_xarray(data, dataset_name, attrs, channel)

    @staticmethod
    def _get_channel_and_measurand(dataset_name):
        channel = measurand = dataset_name
        if dataset_name == "longitude":
            channel = measurand = "lon"
        elif dataset_name == "latitude":
            channel = measurand = "lat"
        return channel, measurand

    def _read_channel_coordinates(self, channel):
        if channel not in self.lons:
            self.lons[channel] = self.h5f["lon"]
            self.lats[channel] = self.h5f["lat"]

    def available_datasets(self, configured_datasets=None):
        """Form the names for the available datasets."""
        for dset in POST_V2025_DATASETS:
            dset_name = dset
            if dset == "lon":
                dset_name = "longitude"
            elif dset == "lat":
                dset_name = "latitude"
            ds_info = {
                "file_type": self.filetype_info["file_type"],
                "resolution": self.resolution,
                "name": dset_name,
            }
            yield True, ds_info

    def _get_period(self, channel):
        if self.period is None:
            interval = self.h5f.attrs["sampling_interval"].astype(str).item()
            self.period = float(interval.split()[0])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Composite classes for the GLM instrument."""

import logging

import xarray as xr

from satpy.enhancements.enhancer import get_enhanced_image

from .core import GenericCompositor

LOG = logging.getLogger(__name__)


class HighlightCompositor(GenericCompositor):
    """Highlight pixels of a layer by an amount determined by a secondary layer.

    The highlighting is applied per channel to either add or subtract an
    intensity from the primary image. In the addition case, the code is
    essentially doing::

        highlight_factor = (highlight_data - min_highlight) / (max_highlight - min_highlight)
        channel_result = primary_data + highlight_factor * max_factor

    The ``max_factor`` is defined per channel and can be positive for an
    additive effect, negative for a subtractive effect, or zero for no
    effect.

    """

    def __init__(self, name, min_highlight=0.0, max_highlight=10.0,  # noqa: D417
                 max_factor=(0.8, 0.8, -0.8, 0), **kwargs):
        """Initialize composite with highlight factor options.

        Args:
            min_highlight (float): Minimum raw value of the "highlight" data
                that will be used for linearly scaling the data along with
                ``max_highlight``.
            max_highlight (float): Maximum raw value of the "highlight" data
                that will be used for linearly scaling the data along with
                ``min_highlight``.
            max_factor (tuple): Maximum effect that the highlight data can
                have on each channel of the primary image data. This will be
                multiplied by the linearly scaled highlight data and then
                added or subtracted from the highlight channels. See class
                docstring for more information. By default this is set to
                ``(0.8, 0.8, -0.8, 0)`` meaning the Red and Green channel
                will be added to by at most 0.8, the Blue channel will be
                subtracted from by at most 0.8 (resulting in yellow highlights),
                and the Alpha channel will not be affected.

        """
        self.min_highlight = min_highlight
        self.max_highlight = max_highlight
        self.max_factor = max_factor
        super().__init__(name, **kwargs)

    @staticmethod
    def _get_enhanced_background_data(background_layer):
        img = get_enhanced_image(background_layer)
        img.data = img.data.clip(0.0, 1.0)
        img = img.convert("RGBA")
        return img.data

    def _get_highlight_factor(self, highlight_data):
        factor = (highlight_data - self.min_highlight) / (self.max_highlight - self.min_highlight)
        factor = factor.where(factor.notnull(), 0)
        return factor

    def _apply_highlight_effect(self, background_data, factor):
        new_channels = []
        for max_factor, band_name in zip(self.max_factor, "RGBA"):
            new_channel = background_data.sel(bands=[band_name])
            if max_factor != 0 or max_factor is not None:
                new_channel = new_channel + factor * max_factor
            new_channels.append(new_channel)
        return new_channels

    def _update_attrs(self, new_data, background_layer, highlight_layer):
        new_data.attrs = background_layer.attrs.copy()
        new_data.attrs["units"] = 1
        new_sensors = self._get_sensors((highlight_layer, background_layer))
        new_data.attrs.update({
            "sensor": new_sensors,
        })

    def __call__(self, projectables, optional_datasets=None, **attrs):
        """Create RGBA image with highlighted pixels."""
        highlight_product, background_layer = self.match_data_arrays(projectables)
        background_data = self._get_enhanced_background_data(background_layer)

        # Adjust the colors of background by highlight layer
        factor = self._get_highlight_factor(highlight_product)
        new_channels = self._apply_highlight_effect(background_data, factor)
        new_data = xr.concat(new_channels, dim="bands")
        self._update_attrs(new_data, background_layer,
                           highlight_product)
        return super(HighlightCompositor, self).__call__((new_data,), **attrs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2021 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Benchmark FCI.

Benchmarks for reading and processing data from the Meteosat Third Generation
(MTG) Flexible Combined Imager (FCI).  Uses pre-launch simulated test data as
published by EUMETSAT in 2020.

Also includes some benchmarks trying different resamplers.
"""

from __future__ import annotations

import fnmatch
import os

import satpy
import satpy.demo.fci

from .utils import GeoBenchmarks


class FCI(GeoBenchmarks):
    """Benchmark FCI FDHSI test data reading."""

    timeout = 600
    region = "eurol"
    reader = "fci_l1c_nc"
    filenames: list[str] = []

    def setup_cache(self, *args):
        """Fetch the data files."""
        fns = self.get_filenames()
        cnt = len(fns)
        if cnt > 40:
            raise ValueError(f"Expected 41 files, found {cnt:d}")
        if cnt < 40:
            fns = satpy.demo.download_fci_test_data()

    def setup(self, *args):
        """Set location of data files."""
        self.filenames = self.get_filenames()

    def get_filenames(self):
        """Get filenames of FCI test data as already available."""
        p = satpy.demo.fci.get_fci_test_data_dir()
        g = p.glob("UNCOMPRESSED/NOMINAL/*-CHK-BODY-*.nc")
        return [os.fspath(fn) for fn in g]

    def time_create_scene(self, chunk):
        """Time to create a scene."""
        names = self._get_filename_selection(chunk)
        self.create_scene(names)
    time_create_scene.params = ["some", "all"]  # type: ignore
    time_create_scene.param_names = ["channel subset"]  # type: ignore

    def peakmem_create_scene(self, chunk):
        """Peak RAM to create a scene."""
        names = self._get_filename_selection(chunk)
        self.create_scene(names)
    peakmem_create_scene.params = time_create_scene.params  # type: ignore
    peakmem_create_scene.param_names = time_create_scene.param_names  # type: ignore

    def time_load(self, chunk, loadable):
        """Time to create a scene and load one channel or composite."""
        names = self._get_filename_selection(chunk)
        self.load_no_padding(loadable, names)
    time_load.params = (time_create_scene.params,  # type: ignore
                        ["ir_105", "natural_color_raw"])
    time_load.param_names = time_create_scene.param_names + ["dataset"]  # type: ignore

    def peakmem_load(self, chunk, loadable):
        """Peak RAM to create a scene and load one channel or composite."""
        names = self._get_filename_selection(chunk)
        self.load_no_padding(loadable, names)
    peakmem_load.params = time_load.params  # type: ignore
    peakmem_load.param_names = time_load.param_names  # type: ignore

    def time_compute(self, chunk, loadable):
        """Time to create a scene and load and compute one channel."""
        names = self._get_filename_selection(chunk)
        self.compute_channel(loadable, names)
    time_compute.params = time_load.params  # type: ignore
    time_compute.param_names = time_load.param_names  # type: ignore

    def peakmem_compute(self, chunk, loadable):
        """Peak memory for creating a scene and loading and computing one channel."""
        names = self._get_filename_selection(chunk)
        self.compute_channel(loadable, names)
    peakmem_compute.params = time_compute.params  # type: ignore
    peakmem_compute.param_names = time_compute.param_names  # type: ignore

    def time_load_resample_compute(self, chunk, loadable, mode):
        """Time to load all chunks, resample, and compute."""
        names = self._get_filename_selection(chunk)
        self.compute_composite(loadable, mode, self.region, names)
    time_load_resample_compute.params = time_load.params + (  # type: ignore
            ["nearest", "bilinear", "gradient_search"],)
    time_load_resample_compute.param_names = time_load.param_names + ["resampler"]  # type: ignore

    def peakmem_load_resample_compute(self, chunk, loadable, mode):
        """Peak memory to load all chunks, resample, and compute."""
        names = self._get_filename_selection(chunk)
        self.compute_composite(loadable, mode, self.region, names)
    peakmem_load_resample_compute.params = time_load_resample_compute.params  # type: ignore
    peakmem_load_resample_compute.param_names = time_load_resample_compute.param_names  # type: ignore

    def time_load_resample_save(self, chunk, loadable, mode):
        """Time to load all chunks, resample, and save."""
        names = self._get_filename_selection(chunk)
        self.save_composite_as_geotiff(loadable, mode, self.region, names)
    time_load_resample_save.params = time_load_resample_compute.params  # type: ignore
    time_load_resample_save.param_names = time_load_resample_compute.param_names  # type: ignore

    def peakmem_load_resample_save(self, chunk, loadable, mode):
        """Peak memory to load all chunks, resample, and save."""
        names = self._get_filename_selection(chunk)
        self.save_composite_as_geotiff(loadable, mode, self.region, names)
    peakmem_load_resample_save.params = time_load_resample_save.params  # type: ignore
    peakmem_load_resample_save.param_names = time_load_resample_save.param_names  # type: ignore

    def _get_filename_selection(self, selection):
        if selection == "some":
            return fnmatch.filter(self.filenames, "*3[0123].nc")
        if selection == "all":
            return self.filenames
        raise ValueError("Expected selection some or all, got " +
                         selection)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2014-2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Testing of helper functions."""

import bz2
import datetime as dt
import os
from unittest import mock

import dask.array as da
import numpy as np
import numpy.testing
import pyresample.geometry
import pytest
import xarray as xr
from fsspec.implementations.memory import MemoryFile, MemoryFileSystem
from pyproj import CRS

from satpy.readers.core import utils as hf


@pytest.fixture
def data_file(tmp_path):
    """Generate a dummy file with data."""
    filename = tmp_path / "test.DAT"
    with filename.open("wb") as fp:
        fp.write(b"TEST")
    return filename


@pytest.fixture
def bzipped_data_file(tmp_path):
    """Generate a bzipped dummy file with data."""
    filename = tmp_path / "test.DAT.bz2"
    with bz2.open(filename, "wb") as fp:
        fp.write(b"TEST")
    return filename


class TestHelpers:
    """Test the area helpers."""

    def test_lonlat_from_geos(self):
        """Get lonlats from geos."""
        import pyproj
        geos_area = mock.MagicMock()
        lon_0 = 0
        h = 35785831.00
        geos_area.crs = CRS({
            "a": 6378169.00,
            "b": 6356583.80,
            "h": h,
            "lon_0": lon_0,
            "proj": "geos"})

        proj = pyproj.Proj(geos_area.crs)
        expected = proj(0, 0, inverse=True)
        np.testing.assert_allclose(expected,
                                   hf._lonlat_from_geos_angle(0, 0, geos_area))

        expected = proj(0, 1000000, inverse=True)
        np.testing.assert_allclose(expected,
                                   hf._lonlat_from_geos_angle(0, 1000000 / h,
                                                              geos_area))

        expected = proj(1000000, 0, inverse=True)
        np.testing.assert_allclose(expected,
                                   hf._lonlat_from_geos_angle(1000000 / h, 0,
                                                              geos_area))

        expected = proj(2000000, -2000000, inverse=True)
        np.testing.assert_allclose(expected,
                                   hf._lonlat_from_geos_angle(2000000 / h,
                                                              -2000000 / h,
                                                              geos_area))

    def test_get_geostationary_bbox(self):
        """Get the geostationary bbox."""
        geos_area = mock.MagicMock()
        lon_0 = 0
        geos_area.crs = CRS({
            "proj": "geos",
            "lon_0": lon_0,
            "a": 6378169.00,
            "b": 6356583.80,
            "h": 35785831.00,
            "units": "m"})
        geos_area.area_extent = [-5500000., -5500000., 5500000., 5500000.]

        lon, lat = hf.get_geostationary_bounding_box(geos_area, 20)
        elon = np.array([-74.802824, -73.667708, -69.879687, -60.758081,
                         -32.224989, 32.224989, 60.758081, 69.879687,
                         73.667708, 74.802824, 74.802824, 73.667708,
                         69.879687, 60.758081, 32.224989, -32.224989,
                         -60.758081, -69.879687, -73.667708, -74.802824])

        elat = -np.array([-6.81982903e-15, -1.93889346e+01, -3.84764764e+01,
                          -5.67707359e+01, -7.18862588e+01, -7.18862588e+01,
                          -5.67707359e+01, -3.84764764e+01, -1.93889346e+01,
                          0.00000000e+00, 6.81982903e-15, 1.93889346e+01,
                          3.84764764e+01, 5.67707359e+01, 7.18862588e+01,
                          7.18862588e+01, 5.67707359e+01, 3.84764764e+01,
                          1.93889346e+01, -0.00000000e+00])

        np.testing.assert_allclose(lon, elon + lon_0)
        np.testing.assert_allclose(lat, elat)

    def test_get_geostationary_angle_extent(self):
        """Get max geostationary angles."""
        geos_area = mock.MagicMock()
        proj_dict = {
            "proj": "geos",
            "sweep": "x",
            "lon_0": -89.5,
            "a": 6378169.00,
            "b": 6356583.80,
            "h": 35785831.00,
            "units": "m"}
        geos_area.crs = CRS(proj_dict)
        expected = (0.15185342867090912, 0.15133555510297725)
        np.testing.assert_allclose(expected,
                                   hf.get_geostationary_angle_extent(geos_area))

        proj_dict["a"] = 1000.0
        proj_dict["b"] = 1000.0
        proj_dict["h"] = np.sqrt(2) * 1000.0 - 1000.0
        geos_area.reset_mock()
        geos_area.crs = CRS(proj_dict)
        expected = (np.deg2rad(45), np.deg2rad(45))
        np.testing.assert_allclose(expected,
                                   hf.get_geostationary_angle_extent(geos_area))

        proj_dict = {
            "proj": "geos",
            "sweep": "x",
            "lon_0": -89.5,
            "ellps": "GRS80",
            "h": 35785831.00,
            "units": "m"}
        geos_area.crs = CRS(proj_dict)
        expected = (0.15185277703584374, 0.15133971368991794)
        np.testing.assert_allclose(expected,
                                   hf.get_geostationary_angle_extent(geos_area))

    def test_geostationary_mask(self):
        """Test geostationary mask."""
        # Compute mask of a very elliptical earth
        area = pyresample.geometry.AreaDefinition(
            "FLDK",
            "Full Disk",
            "geos",
            {"a": "6378169.0",
             "b": "3000000.0",
             "h": "35785831.0",
             "lon_0": "145.0",
             "proj": "geos",
             "units": "m"},
            101,
            101,
            (-6498000.088960204, -6498000.088960204,
             6502000.089024927, 6502000.089024927))

        mask = hf.get_geostationary_mask(area).astype(int).compute()

        # Check results along a couple of lines
        # a) Horizontal
        assert np.all(mask[50, :8] == 0)
        assert np.all(mask[50, 8:93] == 1)
        assert np.all(mask[50, 93:] == 0)

        # b) Vertical
        assert np.all(mask[:31, 50] == 0)
        assert np.all(mask[31:70, 50] == 1)
        assert np.all(mask[70:, 50] == 0)

        # c) Top left to bottom right
        assert np.all(mask[range(33), range(33)] == 0)
        assert np.all(mask[range(33, 68), range(33, 68)] == 1)
        assert np.all(mask[range(68, 101), range(68, 101)] == 0)

        # d) Bottom left to top right
        assert np.all(mask[range(101 - 1, 68 - 1, -1), range(33)] == 0)
        assert np.all(mask[range(68 - 1, 33 - 1, -1), range(33, 68)] == 1)
        assert np.all(mask[range(33 - 1, -1, -1), range(68, 101)] == 0)

    @mock.patch("satpy.readers.core.utils.AreaDefinition")
    def test_sub_area(self, adef):
        """Sub area slicing."""
        area = mock.MagicMock()
        area.pixel_size_x = 1.5
        area.pixel_size_y = 1.5
        area.pixel_upper_left = (0, 0)
        area.area_id = "fakeid"
        area.name = "fake name"
        area.proj_id = "fakeproj"
        area.crs = "some_crs"

        hf.get_sub_area(area, slice(1, 4), slice(0, 3))
        adef.assert_called_once_with("fakeid", "fake name", "fakeproj",
                                     "some_crs",
                                     3, 3,
                                     (0.75, -3.75, 5.25, 0.75))

    def test_np2str(self):
        """Test the np2str function."""
        # byte object
        npbytes = np.bytes_("hej")
        assert hf.np2str(npbytes) == "hej"

        # single element numpy array
        np_arr = np.array([npbytes])
        assert hf.np2str(np_arr) == "hej"

        # scalar numpy array
        np_arr = np.array(npbytes)
        assert hf.np2str(np_arr) == "hej"

        # multi-element array
        npbytes = np.array([npbytes, npbytes])
        with pytest.raises(ValueError, match="Array is not a string type or is larger than 1"):
            hf.np2str(npbytes)

        # non-array
        with pytest.raises(ValueError, match="Array is not a string type or is larger than 1"):
            hf.np2str(5)

    def test_get_earth_radius(self):
        """Test earth radius computation."""
        a = 2.
        b = 1.

        def re(lat):
            """Compute ellipsoid radius at the given geodetic latitude.

            Reference: Capderou, M.: Handbook of Satellite Orbits, Equation (2.20).
            """
            lat = np.deg2rad(lat)
            e2 = 1 - b ** 2 / a ** 2
            n = a / np.sqrt(1 - e2*np.sin(lat)**2)
            return n * np.sqrt((1 - e2)**2 * np.sin(lat)**2 + np.cos(lat)**2)

        for lon in (0, 180, 270):
            assert hf.get_earth_radius(lon=lon, lat=0.0, a=a, b=b) == a
        for lat in (90, -90):
            assert hf.get_earth_radius(lon=0.0, lat=lat, a=a, b=b) == b
        assert np.isclose(hf.get_earth_radius(lon=123, lat=45.0, a=a, b=b), re(45.0))

    def test_reduce_mda(self):
        """Test metadata size reduction."""
        mda = {"a": 1,
               "b": np.array([1, 2, 3]),
               "c": np.array([1, 2, 3, 4]),
               "d": {"a": 1,
                     "b": np.array([1, 2, 3]),
                     "c": np.array([1, 2, 3, 4]),
                     "d": {"a": 1,
                           "b": np.array([1, 2, 3]),
                           "c": np.array([1, 2, 3, 4])}}}
        exp = {"a": 1,
               "b": np.array([1, 2, 3]),
               "d": {"a": 1,
                     "b": np.array([1, 2, 3]),
                     "d": {"a": 1,
                           "b": np.array([1, 2, 3])}}}
        numpy.testing.assert_equal(hf.reduce_mda(mda, max_size=3), exp)

        # Make sure, reduce_mda() doesn't modify the original dictionary
        assert "c" in mda
        assert "c" in mda["d"]
        assert "c" in mda["d"]["d"]

    @mock.patch("satpy.readers.core.utils.bz2.BZ2File")
    @mock.patch("satpy.readers.core.utils.Popen")
    def test_unzip_file(self, mock_popen, mock_bz2):
        """Test the bz2 file unzipping techniques."""
        process_mock = mock.Mock()
        attrs = {"communicate.return_value": (b"output", b"error"),
                 "returncode": 0}
        process_mock.configure_mock(**attrs)
        mock_popen.return_value = process_mock

        bz2_mock = mock.MagicMock()
        bz2_mock.__enter__.return_value.read.return_value = b"TEST"
        mock_bz2.return_value = bz2_mock

        filename = "tester.DAT.bz2"
        whichstr = "satpy.readers.core.utils.which"
        segment = 3
        segmentstr = str(segment).zfill(2)
        # no pbzip2 installed with prefix
        with mock.patch(whichstr) as whichmock:
            whichmock.return_value = None
            new_fname = hf.unzip_file(filename, prefix=segmentstr)
            assert bz2_mock.__enter__.return_value.read.called
            assert os.path.exists(new_fname)
            assert os.path.split(new_fname)[1][0:2] == segmentstr
            if os.path.exists(new_fname):
                os.remove(new_fname)
        # pbzip2 installed without prefix
        with mock.patch(whichstr) as whichmock:
            whichmock.return_value = "/usr/bin/pbzip2"
            new_fname = hf.unzip_file(filename)
            assert mock_popen.called
            assert os.path.exists(new_fname)
            assert os.path.split(new_fname)[1][0:2] != segmentstr
            if os.path.exists(new_fname):
                os.remove(new_fname)

        filename = "tester.DAT"
        new_fname = hf.unzip_file(filename)
        assert new_fname is None

    def test_generic_open_BZ2File(self, bzipped_data_file):
        """Test the generic_open method with bz2 filename input."""
        with hf.generic_open(bzipped_data_file, "rb") as file_object:
            data = file_object.read()
            assert data == b"TEST"

    def test_generic_open_FSFile_MemoryFileSystem(self):
        """Test the generic_open method with FSFile in MemoryFileSystem."""
        from satpy.readers.core.remote import FSFile

        mem_fs = MemoryFileSystem()
        mem_file = MemoryFile(fs=mem_fs, path="{}test.DAT".format(mem_fs.root_marker), data=b"TEST")
        mem_file.commit()
        fsf = FSFile(mem_file)
        with hf.generic_open(fsf, "rb") as file_object:
            data = file_object.read()
            assert data == b"TEST"

    def test_generic_open_filename(self, data_file):
        """Test the generic_open method with filename (str)."""
        with hf.generic_open(data_file, "rb") as file_object:
            data = file_object.read()
            assert data == b"TEST"


    @mock.patch("bz2.decompress", return_value=b"TEST_DECOMPRESSED")
    def test_unzip_FSFile(self, bz2_mock):
        """Test the FSFile bz2 file unzipping techniques."""
        from satpy.readers.core.remote import FSFile

        mock_bz2_decompress = mock.MagicMock()
        mock_bz2_decompress.return_value = b"TEST_DECOMPRESSED"

        segment = 3
        segmentstr = str(segment).zfill(2)

        # test zipped FSFile unzipped in fly (decompress shouldn't be called)
        mem_fs = MemoryFileSystem()
        mem_file = MemoryFile(fs=mem_fs, path="{}test.DAT.bz2".format(mem_fs.root_marker), data=b"TEST")
        mem_file.commit()
        fsf = FSFile(mem_file)

        new_fname = hf.unzip_file(fsf, prefix=segmentstr)
        mock_bz2_decompress.assert_not_called
        bz2_mock.assert_not_called
        assert os.path.exists(new_fname)
        assert os.path.split(new_fname)[1][0:2] == segmentstr
        if os.path.exists(new_fname):
            os.remove(new_fname)

        # test FSFile without unzipping in fly (decompress should be called)
        mem_file = MemoryFile(fs=mem_fs, path="{}test.DAT.bz2".format(mem_fs.root_marker),
                              data=bytes.fromhex("425A68")+b"TEST")
        mem_file.commit()
        fsf = FSFile(mem_file)

        new_fname = hf.unzip_file(fsf, prefix=segmentstr)
        mock_bz2_decompress.assert_called
        bz2_mock.assert_called
        assert os.path.exists(new_fname)
        assert os.path.split(new_fname)[1][0:2] == segmentstr
        if os.path.exists(new_fname):
            os.remove(new_fname)

    @mock.patch("os.remove")
    @mock.patch("satpy.readers.core.utils.unzip_file", return_value="dummy.txt")
    def test_pro_reading_gets_unzipped_file(self, fake_unzip_file, fake_remove):
        """Test the bz2 file unzipping context manager."""
        filename = "dummy.txt.bz2"
        expected_filename = filename[:-4]

        with hf.unzip_context(filename) as new_filename:
            assert new_filename == expected_filename

        fake_unzip_file.assert_called_with(filename)
        fake_remove.assert_called_with(expected_filename)

    def test_apply_rad_correction(self):
        """Test radiance correction technique using user-supplied coefs."""
        slope = 0.5
        offset = -0.1
        res = hf.apply_rad_correction(1.0, slope, offset)
        np.testing.assert_allclose(2.2, res)

    def test_get_user_calibration_factors(self):
        """Test the retrieval of user-supplied calibration factors."""
        radcor_dict = {"WV063": {"slope": 1.015,
                                 "offset": -0.0556},
                       "IR108": {"slo": 1.015,
                                 "off": -0.0556}}
        # Test that correct values are returned from the dict
        slope, offset = hf.get_user_calibration_factors("WV063", radcor_dict)
        assert slope == 1.015
        assert offset == -0.0556

        # Test that channels not present in dict return 1.0, 0.0
        with pytest.warns(UserWarning, match=".*not supplied coefficients.*"):
            slope, offset = hf.get_user_calibration_factors("IR097", radcor_dict)
        assert slope == 1.0
        assert offset == 0.0

        # Check that incorrect dict keys throw an error
        with pytest.raises(KeyError):
            hf.get_user_calibration_factors("IR108", radcor_dict)


class TestSunEarthDistanceCorrection:
    """Tests for applying Sun-Earth distance correction to reflectance."""

    def setup_method(self):
        """Create input / output arrays for the tests."""
        self.test_date = dt.datetime(2020, 8, 15, 13, 0, 40)

        raw_refl = xr.DataArray(da.from_array([10., 20., 40., 1., 98., 50.]),
                                attrs={"start_time": self.test_date,
                                       "scheduled_time": self.test_date})

        corr_refl = xr.DataArray(da.from_array([
            10.25484833, 20.50969667,
            41.01939333, 1.02548483,
            100.49751367, 51.27424167]),
            attrs={"start_time": self.test_date,
                   "scheduled_time": self.test_date},
        )
        self.raw_refl = raw_refl
        self.corr_refl = corr_refl

    def test_get_utc_time(self):
        """Test the retrieval of scene time from a dataset."""
        # First check correct time is returned with 'start_time'
        tmp_array = self.raw_refl.copy()
        del tmp_array.attrs["scheduled_time"]
        utc_time = hf.get_array_date(tmp_array, None)
        assert utc_time == self.test_date

        # Now check correct time is returned with 'scheduled_time'
        tmp_array = self.raw_refl.copy()
        del tmp_array.attrs["start_time"]
        utc_time = hf.get_array_date(tmp_array, None)
        assert utc_time == self.test_date

        # Now check correct time is returned with utc_date passed
        tmp_array = self.raw_refl.copy()
        new_test_date = dt.datetime(2019, 2, 1, 15, 2, 12)
        utc_time = hf.get_array_date(tmp_array, new_test_date)
        assert utc_time == new_test_date

        # Finally, ensure error is raised if no datetime is available
        tmp_array = self.raw_refl.copy()
        del tmp_array.attrs["scheduled_time"]
        del tmp_array.attrs["start_time"]
        with pytest.raises(KeyError):
            hf.get_array_date(tmp_array, None)

    def test_apply_sunearth_corr(self):
        """Test the correction of reflectances with sun-earth distance."""
        out_refl = hf.apply_earthsun_distance_correction(self.raw_refl)
        np.testing.assert_allclose(out_refl, self.corr_refl)
        assert out_refl.attrs["sun_earth_distance_correction_applied"]
        assert isinstance(out_refl.data, da.Array)

    def test_remove_sunearth_corr(self):
        """Test the removal of the sun-earth distance correction."""
        out_refl = hf.remove_earthsun_distance_correction(self.corr_refl)
        np.testing.assert_allclose(out_refl, self.raw_refl)
        assert not out_refl.attrs["sun_earth_distance_correction_applied"]
        assert isinstance(out_refl.data, da.Array)


@pytest.mark.parametrize(("data", "filename", "mode"),
                         [(b"Hello", "dummy.dat", "b"),
                          ("Hello", "dummy.txt", "t")])
def test_generic_open_binary(tmp_path, data, filename, mode):
    """Test the bz2 file unzipping context manager using dummy binary data."""
    dummy_data = data
    dummy_filename = os.fspath(tmp_path / filename)
    with open(dummy_filename, "w" + mode) as f:
        f.write(dummy_data)

    with hf.generic_open(dummy_filename, "r" + mode) as f:
        read_binary_data = f.read()

    assert read_binary_data == dummy_data

    dummy_filename = os.fspath(tmp_path / (filename + ".bz2"))
    with hf.bz2.open(dummy_filename, "w" + mode) as f:
        f.write(dummy_data)

    with hf.generic_open(dummy_filename, "r" + mode) as f:
        read_binary_data = f.read()

    assert read_binary_data == dummy_data


class TestCalibrationCoefficientPicker:
    """Unit tests for calibration coefficient selection."""

    @pytest.fixture(name="coefs")
    def fixture_coefs(self):
        """Get fake coefficients."""
        return {
            "nominal": {
                "ch1": 1.0,
                "ch2": 2.0,
            },
            "mode1": {
                "ch1": 1.1,
            },
            "mode2": {
                "ch2": 2.2,
            }
        }

    @pytest.mark.parametrize(
        ("wishlist", "expected"),
        [
            (
                None,
                {
                    "ch1": {"coefs": 1.0, "mode": "nominal"},
                    "ch2": {"coefs": 2.0, "mode": "nominal"}
                }
            ),
            (
                    "nominal",
                    {
                        "ch1": {"coefs": 1.0, "mode": "nominal"},
                        "ch2": {"coefs": 2.0, "mode": "nominal"}
                    }
            ),
            (
                {("ch1", "ch2"): "nominal"},
                {
                    "ch1": {"coefs": 1.0, "mode": "nominal"},
                    "ch2": {"coefs": 2.0, "mode": "nominal"}
                }
            ),
            (
                {"ch1": "mode1"},
                {
                    "ch1": {"coefs": 1.1, "mode": "mode1"},
                    "ch2": {"coefs": 2.0, "mode": "nominal"}
                }
            ),
            (
                {"ch1": "mode1", "ch2": "mode2"},
                {
                    "ch1": {"coefs": 1.1, "mode": "mode1"},
                    "ch2": {"coefs": 2.2, "mode": "mode2"}
                }
            ),
            (
                {"ch1": "mode1", "ch2": {"gain": 1}},
                {
                    "ch1": {"coefs": 1.1, "mode": "mode1"},
                    "ch2": {"coefs": {"gain": 1}, "mode": "external"}
                }
            ),
        ]
    )
    def test_get_coefs(self, coefs, wishlist, expected):
        """Test getting calibration coefficients."""
        picker = hf.CalibrationCoefficientPicker(coefs, wishlist)
        coefs = {
            channel: picker.get_coefs(channel)
            for channel in ["ch1", "ch2"]
        }
        assert coefs == expected

    @pytest.mark.parametrize(
        "wishlist", ["foo", {"ch1": "foo"}, {("ch1", "ch2"): "foo"}]
    )
    def test_unknown_mode(self, coefs, wishlist):
        """Test handling of unknown calibration mode."""
        with pytest.raises(KeyError, match="Unknown calibration mode"):
            hf.CalibrationCoefficientPicker(coefs, wishlist)

    @pytest.mark.parametrize(
        "wishlist", ["mode1", {"ch2": "mode1"}, {("ch1", "ch2"): "mode1"}]
    )
    def test_missing_coefs(self, coefs, wishlist):
        """Test that an exception is raised when coefficients are missing."""
        picker = hf.CalibrationCoefficientPicker(coefs, wishlist)
        with pytest.raises(KeyError, match="No mode1 calibration"):
            picker.get_coefs("ch2")

    @pytest.mark.parametrize(
        "wishlist", ["mode1", {"ch2": "mode1"}, {("ch1", "ch2"): "mode1"}]
    )
    def test_fallback_to_nominal(self, coefs, wishlist, caplog):
        """Test falling back to nominal coefficients."""
        picker = hf.CalibrationCoefficientPicker(coefs, wishlist,
                                                 fallback="nominal")
        expected = {"coefs": 2.0, "mode": "nominal"}
        assert picker.get_coefs("ch2") == expected
        assert "Falling back" in caplog.text

    def test_no_default_coefs(self):
        """Test initialization without default coefficients."""
        with pytest.raises(KeyError, match="Need at least"):
            hf.CalibrationCoefficientPicker({}, {})

    def test_no_fallback(self):
        """Test initialization without fallback coefficients."""
        with pytest.raises(KeyError, match="No fallback calibration"):
            hf.CalibrationCoefficientPicker({"nominal": 123}, {}, fallback="foo")

    def test_invalid_wishlist_type(self):
        """Test handling of invalid wishlist type."""
        with pytest.raises(TypeError, match="Unsupported wishlist type"):
            hf.CalibrationCoefficientPicker({"nominal": 123}, 123)


@pytest.mark.parametrize("name",
                         ["np2str",
                          "get_geostationary_angle_extent",
                          "get_geostationary_mask",
                          "get_geostationary_bounding_box",
                          "get_sub_area",
                          "unzip_file",
                          "unzip_context",
                          "generic_open",
                          "fromfile",
                          "bbox",
                          "get_earth_radius",
                          "reduce_mda",
                          "get_user_calibration_factors",
                          "apply_rad_correction",
                          "get_array_date",
                          "apply_earthsun_distance_correction",
                          "remove_earthsun_distance_correction",
                          "CalibrationCoefficientPicker",
                          ]
                         )
def test_init_import_warns(name):
    """Test that importing non-reader functions from __init__.py issue a warning."""
    from satpy.readers import utils

    with pytest.warns(UserWarning, match=".*has been moved.*"):
        _ = getattr(utils, name)
